#
# XML API wrappers
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from .logger import log
from .xmletree import ETreeAPI
from .xmllibxml2 import Libxml2API

_backend = os.environ.get("VIRTINST_XML_BACKEND")
log.debug("VIRTINST_XML_BACKEND=%s", _backend)


def _get_default():  # pragma: no cover
    if _backend == "libxml2":
        return Libxml2API
    elif _backend == "etree":
        return ETreeAPI

    try:
        import libxml2

        _ignore = libxml2
        return Libxml2API
    except ImportError as e:
        log.debug("libxml2 import error: %s", e)
        return ETreeAPI


XMLAPI = _get_default()
log.debug("Using XMLAPI=%s", XMLAPI)



# Copyright (C) 2006-2007, 2013, 2014 Red Hat, Inc.
# Copyright (C) 2006 Hugh O. Brock <hbrock@redhat.com>
# Copyright (C) 2014 SUSE LINUX Products GmbH, Nuernberg, Germany.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gtk

import virtinst

from ..lib import uiutil
from ..baseclass import vmmGObjectUI

_EDIT_GFX_ENUM = range(1, 6)
(
    _EDIT_GFX_TYPE,
    _EDIT_GFX_LISTEN,
    _EDIT_GFX_PORT,
    _EDIT_GFX_OPENGL,
    _EDIT_GFX_PASSWD,
) = _EDIT_GFX_ENUM


class vmmGraphicsDetails(vmmGObjectUI):
    __gsignals__ = {
        "changed": (vmmGObjectUI.RUN_FIRST, None, []),
    }

    def __init__(self, vm, builder, topwin):
        vmmGObjectUI.__init__(self, "gfxdetails.ui", None, builder=builder, topwin=topwin)
        self.vm = vm
        self.conn = vm.conn

        self._active_edits = []

        def _e(edittype):
            def signal_cb(*args):
                self._change_cb(edittype)

            return signal_cb

        self.builder.connect_signals(
            {
                "on_graphics_show_password": self._show_password_cb,
                "on_graphics_port_auto_toggled": self._change_port_auto,
                "on_graphics_opengl_toggled": _e(_EDIT_GFX_OPENGL),
                "on_graphics_type_changed": _e(_EDIT_GFX_TYPE),
                "on_graphics_use_password": _e(_EDIT_GFX_PASSWD),
                "on_graphics_listen_type_changed": _e(_EDIT_GFX_LISTEN),
                "on_graphics_password_changed": _e(_EDIT_GFX_PASSWD),
                "on_graphics_address_changed": _e(_EDIT_GFX_LISTEN),
                "on_graphics_port_changed": _e(_EDIT_GFX_PORT),
                "on_graphics_rendernode_changed": _e(_EDIT_GFX_OPENGL),
            }
        )

        self._init_ui()
        self.top_box = self.widget("graphics-box")

    def _cleanup(self):
        self.vm = None
        self.conn = None

    #####################
    # Pretty UI helpers #
    #####################

    @staticmethod
    def graphics_pretty_type_simple(gtype):
        if gtype in [
            virtinst.DeviceGraphics.TYPE_VNC,
            virtinst.DeviceGraphics.TYPE_SDL,
            virtinst.DeviceGraphics.TYPE_RDP,
        ]:
            return str(gtype).upper()
        return str(gtype).capitalize()

    ##########################
    # Initialization methods #
    ##########################

    def _init_ui(self):
        graphics_list = self.widget("graphics-type")
        graphics_model = Gtk.ListStore(str, str)
        graphics_list.set_model(graphics_model)
        uiutil.init_combo_text_column(graphics_list, 1)
        graphics_model.clear()
        graphics_model.append(["spice", _("Spice server")])
        graphics_model.append(["vnc", _("VNC server")])

        graphics_listen_list = self.widget("graphics-listen-type")
        graphics_listen_model = Gtk.ListStore(str, str)
        graphics_listen_list.set_model(graphics_listen_model)
        uiutil.init_combo_text_column(graphics_listen_list, 1)
        graphics_listen_model.clear()
        graphics_listen_model.append(["address", _("Address")])
        graphics_listen_model.append(["none", _("None")])

        self.widget("graphics-address").set_model(Gtk.ListStore(str, str))
        uiutil.init_combo_text_column(self.widget("graphics-address"), 1)

        model = self.widget("graphics-address").get_model()
        model.clear()
        model.append([None, _("Hypervisor default")])
        model.append(["127.0.0.1", _("Localhost only")])
        model.append(["0.0.0.0", _("All interfaces")])

        # Host GPU rendernode
        combo = self.widget("graphics-rendernode")
        model = Gtk.ListStore(str, str)
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)
        model.append([None, _("Auto")])
        devs = self.conn.filter_nodedevs("drm")
        for i in devs:
            drm = i.xmlobj
            if drm.is_drm_render():
                rendernode = drm.get_devnode().path
                model.append([rendernode, i.pretty_name()])

    ##############
    # UI syncing #
    ##############

    def _sync_ui(self):
        gtype = uiutil.get_list_selection(self.widget("graphics-type"))
        is_vnc = gtype == "vnc"
        is_spice = gtype == "spice"

        listen = uiutil.get_list_selection(self.widget("graphics-listen-type"))
        has_listen_none = listen in ["none", "socket"]

        has_virtio_3d = bool(
            [v for v in self.vm.xmlobj.devices.video if (v.model == "virtio" and v.accel3d)]
        )

        uiutil.set_grid_row_visible(self.widget("graphics-warn-virtio"), not has_virtio_3d)
        uiutil.set_grid_row_visible(self.widget("graphics-warn-listen"), not has_listen_none)

        passwd_enabled = self.widget("graphics-password-chk").get_active()
        self.widget("graphics-password").set_sensitive(passwd_enabled)
        if not passwd_enabled:
            self.widget("graphics-password").set_text("")
        passwd_visible = self.widget("graphics-visibility-chk").get_active()
        self.widget("graphics-password").set_visibility(passwd_visible)

        glval = self.widget("graphics-opengl").get_active()
        uiutil.set_grid_row_visible(self.widget("graphics-opengl-subopts-box"), glval and is_spice)

        all_rows = [
            "graphics-listen-type",
            "graphics-address",
            "graphics-password-box",
            "graphics-port-box",
            "graphics-opengl",
            "graphics-opengl-subopts-box",
        ]

        is_auto = (
            self.widget("graphics-port-auto").get_active()
            or self.widget("graphics-port-auto").get_inconsistent()
        )
        self.widget("graphics-port").set_visible(not is_auto)

        rows = ["graphics-password-box", "graphics-listen-type"]
        if listen == "address":
            rows.extend(["graphics-port-box", "graphics-address"])
        if is_spice:
            rows.append("graphics-opengl")
            if glval:
                rows.append("graphics-opengl-subopts-box")

        if not is_vnc and not is_spice:
            rows = []

        for row in all_rows:
            uiutil.set_grid_row_visible(self.widget(row), row in rows)

    ##############
    # Public API #
    ##############

    def reset_state(self):
        uiutil.set_list_selection(self.widget("graphics-type"), self.config.get_graphics_type())
        self.widget("graphics-listen-type").set_active(0)
        self.widget("graphics-address").set_active(0)

        # Select last entry in the list, which should be a rendernode path
        rendermodel = self.widget("graphics-rendernode").get_model()
        self.widget("graphics-rendernode").set_active_iter(rendermodel[-1].iter)

        self.widget("graphics-port-auto").set_active(True)
        self.widget("graphics-password").set_text("")
        self.widget("graphics-password").set_sensitive(False)
        self.widget("graphics-password-chk").set_active(False)
        self.widget("graphics-opengl").set_active(False)
        self._sync_ui()
        self._active_edits = []

    def set_dev(self, gfx):
        self.reset_state()
        portval = gfx.port
        portautolabel = _("A_uto")

        if portval == -1 or gfx.autoport:
            portauto = True
            if portval and portval != -1:  # pragma: no cover
                # Triggering this with the test driver is tough
                # because it doesn't fill in runtime port values
                portautolabel = _("A_uto (Port %(port)d)") % {"port": portval}
        elif portval is None:
            portauto = None
        else:
            portauto = False

        self.widget("graphics-port").set_value(portval or 0)
        self.widget("graphics-port-auto").set_label(portautolabel)
        self.widget("graphics-port-auto").set_active(bool(portauto))
        self.widget("graphics-port-auto").set_inconsistent(portauto is None)

        gtype = gfx.type
        uiutil.set_list_selection(self.widget("graphics-type"), gtype)

        use_passwd = gfx.passwd is not None
        self.widget("graphics-password-chk").set_active(use_passwd)
        self.widget("graphics-password").set_text(gfx.passwd or "")

        listentype = gfx.get_first_listen_type()
        uiutil.set_list_selection(self.widget("graphics-listen-type"), listentype)
        uiutil.set_list_selection(self.widget("graphics-address"), gfx.listen)

        glval = bool(gfx.gl)
        renderval = gfx.rendernode or None
        self.widget("graphics-opengl").set_active(glval)

        if glval:
            # Only sync rendernode UI with XML, if gl=on, otherwise
            # we want to preserve the suggested rendernode already
            # selected in the UI
            uiutil.set_list_selection(self.widget("graphics-rendernode"), renderval)

        # This comes last
        self._active_edits = []

    def get_values(self):
        ret = {}

        glval = self.widget("graphics-opengl").get_active()
        if not self.widget("graphics-opengl").is_visible():
            glval = None

        rendernode = uiutil.get_list_selection(self.widget("graphics-rendernode"))
        if not self.widget("graphics-rendernode").is_visible():
            rendernode = None

        passwd = self.widget("graphics-password").get_text()
        if not self.widget("graphics-password-chk").get_active():
            passwd = None

        portauto = self.widget("graphics-port-auto").get_active()
        port = uiutil.spin_get_helper(self.widget("graphics-port"))

        if (
            self.widget("graphics-port-auto").get_inconsistent()
            and self.widget("graphics-port-auto").is_visible()
        ):
            # When switching from listen=None to listen=address with
            # Hypervisor default, we need to force autoport otherwise
            # the XML change doesn't stick
            self._active_edits.append(_EDIT_GFX_PORT)
            portauto = True
        if portauto:
            port = -1

        listen = uiutil.get_list_selection(self.widget("graphics-listen-type"))
        addr = uiutil.get_list_selection(self.widget("graphics-address"))
        if listen and listen == "none":
            port = None
        elif listen:
            listen = addr

        gtype = uiutil.get_list_selection(self.widget("graphics-type"))

        if _EDIT_GFX_TYPE in self._active_edits:
            ret["gtype"] = gtype
        if _EDIT_GFX_PORT in self._active_edits:
            ret["port"] = port
        if _EDIT_GFX_LISTEN in self._active_edits:
            ret["listen"] = listen
        if _EDIT_GFX_PASSWD in self._active_edits:
            ret["passwd"] = passwd
        if _EDIT_GFX_OPENGL in self._active_edits:
            ret["gl"] = glval
            ret["rendernode"] = rendernode

        return ret

    def build_device(self):
        self._active_edits = _EDIT_GFX_ENUM[:]
        values = self.get_values()
        dev = virtinst.DeviceGraphics(self.conn.get_backend())

        dev.type = values["gtype"]
        dev.passwd = values["passwd"]
        dev.gl = values["gl"]
        dev.rendernode = values["rendernode"]
        dev.listen = values["listen"]
        dev.port = values["port"]

        return dev

    #############
    # Listeners #
    #############

    def _change_cb(self, edittype):
        self._sync_ui()
        if edittype not in self._active_edits:
            self._active_edits.append(edittype)
        self.emit("changed")

    def _change_port_auto(self, ignore):
        self.widget("graphics-port-auto").set_inconsistent(False)
        self._change_cb(_EDIT_GFX_PORT)

    def _show_password_cb(self, src):
        self._sync_ui()



#
# Storage lookup/creation helpers
#
# Copyright 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import re
import stat
import subprocess

import libvirt

from .logger import log
from .storage import StoragePool, StorageVolume
from . import xmlutil


def _lookup_vol_by_path(conn, path):
    """
    Try to find a volume matching the full passed path. Call info() on
    it to ensure the volume wasn't removed behind libvirt's back
    """
    try:
        vol = conn.storageVolLookupByPath(path)
        vol.info()
        return vol, None
    except libvirt.libvirtError as e:
        # test_urls trigger empty errors here, because python
        # garbage collection kicks in after the failure but before
        # we read the error code, and libvirt virStoragePoolFree
        # public entry point clears the cached error. So ignore
        # an empty error code
        if e.get_error_code() and e.get_error_code() != libvirt.VIR_ERR_NO_STORAGE_VOL:
            raise  # pragma: no cover
        return None, e


def _lookup_vol_by_basename(pool, path):
    """
    Try to lookup a volume for 'path' in parent 'pool' by it's filename.
    This sometimes works in cases where full volume path lookup doesn't,
    since not all libvirt storage backends implement path lookup.
    """
    name = os.path.basename(path)
    if name in pool.listVolumes():
        return pool.storageVolLookupByName(name)


def _get_block_size(path):  # pragma: no cover
    try:
        fd = os.open(path, os.O_RDONLY)
        # os.SEEK_END is not present on all systems
        size = os.lseek(fd, 0, 2)
        os.close(fd)
    except Exception:
        size = 0
    return size


def _get_size(path):
    if not os.path.exists(path):
        return 0
    if _stat_is_block(path):
        return _get_block_size(path)  # pragma: no cover
    return os.path.getsize(path)


def _stat_is_block(path):
    if not os.path.exists(path):
        return False
    return stat.S_ISBLK(os.stat(path)[stat.ST_MODE])


def _check_if_path_managed(conn, path):
    """
    Try to lookup storage objects for the passed path.

    Returns (volume, parent pool). Only one is returned at a time.
    """
    vol, ignore = _lookup_vol_by_path(conn, path)
    if vol:
        return vol, vol.storagePoolLookupByVolume()

    pool = StoragePool.lookup_pool_by_path(conn, os.path.dirname(path))
    if not pool:
        return None, None

    # We have the parent pool, but didn't find a volume on first lookup
    # attempt. Refresh the pool and try again, in case we were just out
    # of date or the pool was inactive.
    try:
        StoragePool.ensure_pool_is_running(pool, refresh=True)
        vol, verr = _lookup_vol_by_path(conn, path)
        if verr:
            try:
                vol = _lookup_vol_by_basename(pool, path)
            except Exception:  # pragma: no cover
                pass
    except Exception as e:  # pragma: no cover
        vol = None
        pool = None
        verr = str(e)

    if not vol and not pool and verr:  # pragma: no cover
        raise ValueError(_("Cannot use storage %(path)s: %(err)s") % {"path": path, "err": verr})

    return vol, pool


def _can_auto_manage(path):
    path = path or ""
    skip_prefixes = ["/dev", "/sys", "/proc"]

    if path_is_url(path):
        return False

    for prefix in skip_prefixes:
        if path.startswith(prefix + "/") or path == prefix:
            return False
    return True


def _get_storage_search_path(path):
    # If the passed path is one of our artificial rbd:// style
    # URIs, parse out the path component, since that is what is needed
    # for looking up storage volumes by target path
    from .uri import URI

    uriobj = URI(path)
    if uriobj.scheme == "rbd":
        return uriobj.path.strip("/")
    return path


def manage_path(conn, path):
    """
    If path is not managed, try to create a storage pool to probe the path
    """
    if not conn.support.conn_storage():
        return None, None, None  # pragma: no cover
    if not path:
        return None, None, None

    if not path_is_url(path) and not path_is_network_vol(conn, path):
        path = os.path.abspath(path)

    searchpath = _get_storage_search_path(path)
    vol, pool = _check_if_path_managed(conn, searchpath)
    if vol or pool or not _can_auto_manage(path):
        return path, vol, pool

    dirname = os.path.dirname(path)
    poolname = os.path.basename(dirname).replace(" ", "_")
    if not poolname:
        poolname = "dirpool"
    poolname = StoragePool.find_free_name(conn, poolname)
    log.debug("Attempting to build pool=%s target=%s", poolname, dirname)

    poolxml = StoragePool(conn)
    poolxml.name = poolname
    poolxml.type = poolxml.TYPE_DIR
    poolxml.target_path = dirname
    pool = poolxml.install(build=False, create=True, autostart=True)

    vol = _lookup_vol_by_basename(pool, path)
    return path, vol, pool


def path_is_url(path):
    """
    Detect if path is a URL
    """
    return bool(re.match(r"[a-zA-Z]+(\+[a-zA-Z]+)?://.*", path or ""))


def path_is_network_vol(conn, path):
    """
    Detect if path is a network volume such as rbd, gluster, etc
    """
    for volxml in conn.fetch_all_vols():
        if path and volxml.target_path == path:
            return volxml.type == "network"
    return False


def _get_dev_type(path, vol_xml, vol_object, pool_xml, remote):
    """
    Try to get device type for volume.
    """
    if vol_xml and vol_xml.type:
        return vol_xml.type

    if pool_xml:
        t = pool_xml.get_disk_type()
        if t == StorageVolume.TYPE_BLOCK:
            return "block"
        elif t == StorageVolume.TYPE_NETWORK:
            return "network"

    if vol_object:  # pragma: no cover
        # This path is hard to test, because test suite XML always has
        # the vol_xml.type set
        t = vol_object.info()[0]
        if t == StorageVolume.TYPE_FILE:
            return "file"
        elif t == StorageVolume.TYPE_BLOCK:
            return "block"
        elif t == StorageVolume.TYPE_NETWORK:
            return "network"

    if path:
        if path_is_url(path):
            return "network"

        if remote:
            if not _can_auto_manage(path):
                # Just a heurisitic, if this path is one of the ones
                # we don't try to auto-import, then consider it a
                # block device, because managing those correctly is difficult
                return "block"

        else:
            if os.path.isdir(path):
                return "dir"
            elif _stat_is_block(path):
                return "block"  # pragma: no cover

    return "file"


def path_definitely_exists(conn, path):
    """
    Return True if the path certainly exists, False if we are unsure.
    See DeviceDisk entry point for more details
    """
    if path is None:
        return False

    try:
        (vol, pool) = _check_if_path_managed(conn, path)
        ignore = pool
        if vol:
            return True

        if not conn.is_remote():
            return os.path.exists(path)
    except Exception:  # pragma: no cover
        pass

    return False


#########################
# ACL/path perm helpers #
#########################

SETFACL = "setfacl"


def _fix_perms_acl(dirname, username):
    cmd = [SETFACL, "--modify", "user:%s:x" % username, dirname]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()

    log.debug("Ran command '%s'", cmd)
    if out or err:
        log.debug("out=%s\nerr=%s", out, err)

    if proc.returncode != 0:
        raise ValueError(err)


def _fix_perms_chmod(dirname):
    log.debug("Setting +x on %s", dirname)
    mode = os.stat(dirname).st_mode
    newmode = mode | stat.S_IXOTH
    os.chmod(dirname, newmode)
    if os.stat(dirname).st_mode != newmode:
        # Trying to change perms on vfat at least doesn't work
        # but also doesn't seem to error. Try and detect that
        raise ValueError(_("Permissions on '%s' did not stick") % dirname)  # pragma: no cover


def set_dirs_searchable(dirlist, username):
    useacl = True
    errdict = {}
    for dirname in dirlist:
        if useacl:
            try:
                _fix_perms_acl(dirname, username)
                continue
            except Exception as e:
                log.debug("setfacl failed: %s", e)
                log.debug("trying chmod")
                useacl = False

        try:
            # If we reach here, ACL setting failed, try chmod
            _fix_perms_chmod(dirname)
        except Exception as e:  # pragma: no cover
            errdict[dirname] = str(e)

    return errdict


def _is_dir_searchable(dirname, uid, username):
    """
    Check if passed directory is searchable by uid
    """
    try:
        statinfo = os.stat(dirname)
    except OSError:  # pragma: no cover
        return False

    if uid == statinfo.st_uid:
        flag = stat.S_IXUSR
    elif uid == statinfo.st_gid:
        flag = stat.S_IXGRP  # pragma: no cover
    else:
        flag = stat.S_IXOTH

    if bool(statinfo.st_mode & flag):
        return True

    # Check POSIX ACL (since that is what we use to 'fix' access)
    cmd = ["getfacl", dirname]
    try:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = proc.communicate()
    except OSError:  # pragma: no cover
        log.debug("Didn't find the getfacl command.")
        return False

    if proc.returncode != 0:  # pragma: no cover
        log.debug("Cmd '%s' failed: %s", cmd, err)
        return False

    pattern = "user:%s:..x" % username
    return bool(re.search(pattern.encode("utf-8", "replace"), out))


def is_path_searchable(path, uid, username):
    """
    Check each dir component of the passed path, see if they are
    searchable by the uid/username, and return a list of paths
    which aren't searchable
    """
    if os.path.isdir(path):
        dirname = path
        base = "-"
    else:
        dirname, base = os.path.split(path)

    fixlist = []
    while base:
        if not _is_dir_searchable(dirname, uid, username):
            fixlist.append(dirname)
        dirname, base = os.path.split(dirname)

    return fixlist


##############################################
# Classes for tracking storage media details #
##############################################


class _StorageBase:
    """
    Storage base class, defining the API used by DeviceDisk
    """

    def __init__(self, conn):
        self._conn = conn
        self._parent_pool_xml = None

    def get_size(self):
        raise NotImplementedError()

    def get_dev_type(self):
        raise NotImplementedError()

    def get_driver_type(self):
        raise NotImplementedError()

    def get_vol_install(self):
        raise NotImplementedError()

    def get_vol_object(self):
        raise NotImplementedError()

    def get_parent_pool(self):
        raise NotImplementedError()

    def get_parent_pool_xml(self):
        if not self._parent_pool_xml and self.get_parent_pool():
            self._parent_pool_xml = StoragePool(
                self._conn, parsexml=self.get_parent_pool().XMLDesc(0)
            )
        return self._parent_pool_xml

    def validate(self):
        raise NotImplementedError()

    def get_path(self):
        raise NotImplementedError()

    def is_stub(self):
        return False

    # Storage creation routines
    def is_size_conflict(self):
        raise NotImplementedError()

    def will_create_storage(self):
        raise NotImplementedError()

    def create(self, meter):
        ignore = meter  # pragma: no cover
        raise xmlutil.DevError("%s can't create storage" % self.__class__.__name__)


class _StorageCreator(_StorageBase):
    """
    Base object for classes that will actually create storage on disk
    """

    def __init__(self, conn):
        _StorageBase.__init__(self, conn)

        self._pool = None
        self._vol_install = None
        self._path = None
        self._size = None
        self._dev_type = None

    ##############
    # Public API #
    ##############

    def create(self, meter):
        raise NotImplementedError

    def validate(self):
        raise NotImplementedError

    def get_size(self):
        raise NotImplementedError

    def get_path(self):
        if self._vol_install and not self._path:
            xmlobj = StoragePool(self._conn, parsexml=self._vol_install.pool.XMLDesc(0))
            if self.get_dev_type() == "network":
                self._path = self._vol_install.name
            else:
                self._path = os.path.join(xmlobj.target_path, self._vol_install.name)
        return self._path

    def get_vol_install(self):
        return self._vol_install

    def get_vol_xml(self):
        return self._vol_install

    def get_dev_type(self):
        if not self._dev_type:
            self._dev_type = _get_dev_type(
                self._path,
                self._vol_install,
                None,
                self.get_parent_pool_xml(),
                self._conn.is_remote(),
            )
        return self._dev_type

    def get_driver_type(self):
        if self._vol_install:
            if self._vol_install.supports_format():
                return self._vol_install.format
        return "raw"

    def will_create_storage(self):
        return True

    def get_vol_object(self):
        return None

    def get_parent_pool(self):
        if self._vol_install:
            return self._vol_install.pool
        return None

    def exists(self):
        return False


class ManagedStorageCreator(_StorageCreator):
    """
    Handles storage creation via libvirt APIs. All the actual creation
    logic lives in StorageVolume, this is mostly about pulling out bits
    from that class and mapping them to DeviceDisk elements
    """

    def __init__(self, conn, vol_install):
        _StorageCreator.__init__(self, conn)

        self._pool = vol_install.pool
        self._vol_install = vol_install

    def create(self, meter):
        return self._vol_install.install(meter=meter)

    def is_size_conflict(self):
        return self._vol_install.is_size_conflict()

    def validate(self):
        return self._vol_install.validate()

    def get_size(self):
        return float(self._vol_install.capacity) / 1024.0 / 1024.0 / 1024.0


class CloneStorageCreator(_StorageCreator):
    """
    Handles manually copying local files for Cloner

    Many clone scenarios will use libvirt storage APIs, which will use
    the ManagedStorageCreator
    """

    def __init__(self, conn, output_path, input_path, size, sparse):
        _StorageCreator.__init__(self, conn)

        self._path = output_path
        self._output_path = output_path
        self._input_path = input_path
        self._size = size
        self._sparse = sparse

    def get_size(self):
        return self._size

    def is_size_conflict(self):
        ret = False
        msg = None
        if self.get_dev_type() == "block":
            avail = _get_size(self._path)  # pragma: no cover
        else:
            vfs = os.statvfs(os.path.dirname(os.path.abspath(self._path)))
            avail = vfs.f_frsize * vfs.f_bavail
        need = int(self._size) * 1024 * 1024 * 1024
        if need > avail:  # pragma: no cover
            if self._sparse:
                msg = _(
                    "The filesystem will not have enough free space"
                    " to fully allocate the sparse file when the guest"
                    " is running."
                )
            else:
                ret = True
                msg = _("There is not enough free space to create the disk.")

            if msg:
                msg += " "
                msg += _("%(mem1)s M requested > %(mem2)s M available") % {
                    "mem1": (need // (1024 * 1024)),
                    "mem2": (avail // (1024 * 1024)),
                }
        return (ret, msg)

    def validate(self):
        if self._size is None:  # pragma: no cover
            raise ValueError(_("size is required for non-existent disk '%s'") % self.get_path())

        err, msg = self.is_size_conflict()
        if err:
            raise ValueError(msg)  # pragma: no cover
        if msg:
            log.warning(msg)  # pragma: no cover

    def create(self, meter):
        text = _("Cloning %(srcfile)s") % {"srcfile": os.path.basename(self._input_path)}

        size_bytes = int(self.get_size() * 1024 * 1024 * 1024)
        meter.start(text, size_bytes)

        # Plain file clone
        self._clone_local(meter, size_bytes)

    def _clone_local(self, meter, size_bytes):
        if self._input_path == "/dev/null":  # pragma: no cover
            # Not really sure why this check is here,
            # but keeping for compat
            log.debug("Source dev was /dev/null. Skipping")
            return
        if self._input_path == self._output_path:
            log.debug("Source and destination are the same. Skipping.")
            return

        # If a destination file exists and sparse flag is True,
        # this priority takes an existing file.

        if not os.path.exists(self._output_path) and self._sparse:
            clone_block_size = 4096
            sparse = True
            fd = None
            try:
                fd = os.open(self._output_path, os.O_WRONLY | os.O_CREAT, 0o640)
                os.ftruncate(fd, size_bytes)
            finally:
                if fd:
                    os.close(fd)
        else:
            clone_block_size = 1024 * 1024 * 10
            sparse = False

        log.debug(
            "Local Cloning %s to %s, sparse=%s, block_size=%s",
            self._input_path,
            self._output_path,
            sparse,
            clone_block_size,
        )

        zeros = b"\0" * 4096

        src_fd, dst_fd = None, None
        try:
            try:
                src_fd = os.open(self._input_path, os.O_RDONLY)
                dst_fd = os.open(self._output_path, os.O_WRONLY | os.O_CREAT, 0o640)

                i = 0
                while 1:
                    data = os.read(src_fd, clone_block_size)
                    s = len(data)
                    if s == 0:
                        meter.end()
                        break
                    # check sequence of zeros
                    if sparse and zeros == data:
                        os.lseek(dst_fd, s, 1)
                    else:
                        b = os.write(dst_fd, data)
                        if s != b:  # pragma: no cover
                            meter.end()
                            break
                    i += s
                    if i < size_bytes:
                        meter.update(i)
            except OSError as e:  # pragma: no cover
                log.debug("Error while cloning", exc_info=True)
                msg = _("Error cloning diskimage %(inputpath)s to %(outputpath)s: %(error)s") % {
                    "inputpath": self._input_path,
                    "outputpath": self._output_path,
                    "error": str(e),
                }
                raise RuntimeError(msg) from None
        finally:
            if src_fd is not None:
                os.close(src_fd)
            if dst_fd is not None:
                os.close(dst_fd)


class StorageBackendStub(_StorageBase):
    """
    Class representing a storage path for a parsed XML disk, that we
    don't want to do slow resolving of unless requested
    """

    def __init__(self, conn, path, dev_type, driver_type):
        _StorageBase.__init__(self, conn)
        self._path = path
        self._dev_type = dev_type
        self._driver_type = driver_type

    def get_path(self):
        return self._path

    def get_vol_object(self):
        return None

    def get_vol_xml(self):
        return None

    def get_parent_pool(self):
        return None

    def get_size(self):
        return 0

    def exists(self):
        return True

    def get_dev_type(self):
        return self._dev_type

    def get_driver_type(self):
        return self._driver_type

    def validate(self):
        return

    def get_vol_install(self):
        return None

    def is_size_conflict(self):
        return (False, None)

    def is_stub(self):
        return True

    def will_create_storage(self):
        return False


class StorageBackend(_StorageBase):
    """
    Class that carries all the info about any existing storage that
    the disk references
    """

    def __init__(self, conn, path, vol_object, parent_pool):
        _StorageBase.__init__(self, conn)

        self._vol_object = vol_object
        self._parent_pool = parent_pool
        self._path = path

        if self._vol_object is not None:
            self._path = None

        if self._vol_object and not self._parent_pool:
            raise xmlutil.DevError("parent_pool must be specified")

        # Cached bits
        self._vol_xml = None
        self._parent_pool_xml = None
        self._exists = None
        self._size = None
        self._dev_type = None

    ##############
    # Public API #
    ##############

    def get_path(self):
        if self._vol_object:
            return self.get_vol_xml().target_path
        return self._path

    def get_vol_object(self):
        return self._vol_object

    def get_vol_xml(self):
        if self._vol_xml is None:
            self._vol_xml = StorageVolume(self._conn, parsexml=self._vol_object.XMLDesc(0))
            self._vol_xml.pool = self._parent_pool
        return self._vol_xml

    def get_parent_pool(self):
        return self._parent_pool

    def get_size(self):
        """
        Return size of existing storage
        """
        if self._size is None:
            ret = 0
            if self._vol_object:
                ret = self.get_vol_xml().capacity
            elif self._path:
                ret = _get_size(self._path)
            self._size = float(ret) / 1024.0 / 1024.0 / 1024.0
        return self._size

    def exists(self):
        if self._exists is None:
            if self._vol_object:
                self._exists = True
            elif self._path is None:
                self._exists = True
            elif (
                self.get_dev_type() != "network"
                and not self._conn.is_remote()
                and os.path.exists(self._path)
            ):
                self._exists = True
            elif self._parent_pool:
                self._exists = False
            elif self.get_dev_type() == "network":
                self._exists = True
            elif self._conn.is_remote() and not _can_auto_manage(self._path):
                # This allows users to pass /dev/sdX and we don't try to
                # validate it exists on the remote connection, since
                # autopooling /dev is perilous. Libvirt will error if
                # the device doesn't exist.
                self._exists = True
            else:
                self._exists = False
        return self._exists

    def get_dev_type(self):
        """
        Return disk 'type' value per storage settings
        """
        if self._dev_type is None:
            vol_xml = None
            if self._vol_object:
                vol_xml = self.get_vol_xml()
            self._dev_type = _get_dev_type(
                self._path,
                vol_xml,
                self._vol_object,
                self.get_parent_pool_xml(),
                self._conn.is_remote(),
            )
        return self._dev_type

    def get_driver_type(self):
        if self._vol_object:
            if self.get_vol_xml().supports_format():
                return self.get_vol_xml().format
            return "raw"
        return None

    def validate(self):
        return

    def get_vol_install(self):
        return None

    def is_size_conflict(self):
        return (False, None)

    def will_create_storage(self):
        return False



# Copyright (C) 2006, 2013-2014 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import queue
import threading

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from virtinst import log

from .baseclass import vmmGObject
from .createconn import vmmCreateConn
from .connmanager import vmmConnectionManager
from .lib.inspection import vmmInspection
from .systray import vmmSystray

(PRIO_HIGH, PRIO_LOW) = range(1, 3)


def _show_startup_error(fn):
    """
    Decorator to show a modal error dialog if an exception is raised
    from a startup routine
    """

    # pylint: disable=protected-access
    def newfn(self, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except Exception as e:
            modal = self._can_exit()
            self.err.show_err(str(e), modal=modal)
            self._exit_app_if_no_windows()

    return newfn


class vmmEngine(vmmGObject):
    CLI_SHOW_MANAGER = "manager"
    CLI_SHOW_DOMAIN_CREATOR = "creator"
    CLI_SHOW_DOMAIN_EDITOR = "editor"
    CLI_SHOW_DOMAIN_PERFORMANCE = "performance"
    CLI_SHOW_DOMAIN_CONSOLE = "console"
    CLI_SHOW_DOMAIN_DELETE = "delete"
    CLI_SHOW_HOST_SUMMARY = "summary"
    CLI_SHOW_SYSTEM_TRAY = "systray"

    @classmethod
    def get_instance(cls):
        if not cls._instance:
            cls._instance = vmmEngine()
        return cls._instance

    __gsignals__ = {
        "app-closing": (vmmGObject.RUN_FIRST, None, []),
    }

    def __init__(self):
        vmmGObject.__init__(self)

        self._exiting = False

        self._window_count = 0
        self._gtkapplication = None
        self._init_gtk_application()

        self._timer = None
        self._tick_counter = 0
        self._tick_thread_slow = False
        self._tick_thread = threading.Thread(
            name="Tick thread", target=self._handle_tick_queue, args=()
        )
        self._tick_thread.daemon = True
        self._tick_queue = queue.PriorityQueue(100)

    @property
    def _connobjs(self):
        return vmmConnectionManager.get_instance().conns

    def _cleanup(self):
        # self._timer should be automatically cleaned up
        pass

    #################
    # init handling #
    #################

    def _default_startup(self, skip_autostart, cliuri):
        """
        Actual startup routines if we are running a new instance of the app
        """
        vmmSystray.get_instance()
        vmmInspection.get_instance()

        self.add_gsettings_handle(
            self.config.on_stats_update_interval_changed(self._timer_changed_cb)
        )

        self._schedule_timer()
        self._tick_thread.start()
        self._tick()

        uris = list(self._connobjs.keys())
        if not uris:
            log.debug("No stored URIs found.")
        else:
            log.debug("Loading stored URIs:\n%s", "  \n".join(sorted(uris)))

        if not skip_autostart:
            self.idle_add(self._autostart_conns)

        if not self.config.get_conn_uris() and not cliuri:
            # Only add default if no connections are currently known
            manager = self._get_manager()
            manager.set_startup_error(_("Checking for virtualization packages..."))
            self.timeout_add(1000, self._add_default_conn)

    def _add_default_conn(self):
        """
        If there's no cached connections, or any requested on the command
        line, try to determine a default URI and open it, first checking
        if libvirt is running
        """
        from .lib import connectauth

        detected_uri = vmmCreateConn.default_uri()
        log.debug("Probed default URI=%s", detected_uri)
        if self.config.CLITestOptions.firstrun_uri is not None:
            detected_uri = self.config.CLITestOptions.firstrun_uri or None
            log.debug("Using test-options firstrun_uri=%s", detected_uri)

        manager = self._get_manager()
        msg = connectauth.setup_first_uri(self.config, detected_uri)
        if msg:
            manager.set_startup_error(msg)
            return

        # Launch idle callback to connect to default URI
        def idle_connect():
            def _open_completed(c, ConnectError):
                if ConnectError:
                    self._handle_conn_error(c, ConnectError)

            conn = vmmConnectionManager.get_instance().add_conn(detected_uri)
            conn.set_autoconnect(True)
            conn.connect_once("open-completed", _open_completed)
            conn.open()

        self.idle_add(idle_connect)

    def _autostart_conns(self):
        """
        We serialize conn autostart, so polkit/ssh-askpass doesn't spam
        """
        if self._exiting:
            return  # pragma: no cover

        connections_queue = queue.Queue()
        auto_conns = [conn.get_uri() for conn in self._connobjs.values() if conn.get_autoconnect()]

        def add_next_to_queue():
            if not auto_conns:
                connections_queue.put(None)
            else:
                connections_queue.put(auto_conns.pop(0))

        def conn_open_completed(_conn, ConnectError):
            # Explicitly ignore connection errors, we've done that
            # for a while and it can be noisy
            if ConnectError is not None:  # pragma: no cover
                log.debug("Autostart connection error: %s", ConnectError.details)
            add_next_to_queue()

        def handle_queue():
            while True:
                uri = connections_queue.get()
                if uri is None:
                    return
                if self._exiting:
                    return  # pragma: no cover
                if uri not in self._connobjs:  # pragma: no cover
                    add_next_to_queue()
                    continue

                conn = self._connobjs[uri]
                conn.connect_once("open-completed", conn_open_completed)
                self.idle_add(conn.open)

        add_next_to_queue()
        self._start_thread(handle_queue, "Conn autostart thread")

    ############################
    # Gtk Application handling #
    ############################

    def _on_gtk_application_activated(self, ignore):
        """
        Invoked after application.run()
        """
        if not self._application.get_windows():
            log.debug("Initial gtkapplication activated")
            self._application.add_window(Gtk.Window())

    def _init_gtk_application(self):
        self._application = Gtk.Application(application_id="org.virt-manager.virt-manager", flags=0)
        self._application.register(None)
        self._application.connect("activate", self._on_gtk_application_activated)

        # pylint: disable=no-member
        action = Gio.SimpleAction.new("cli_command", GLib.VariantType.new("(sss)"))
        action.connect("activate", self._handle_cli_command)
        self._application.add_action(action)

    def start(self, uri, show_window, domain, skip_autostart):
        """
        Public entrypoint from virt-manager cli. If app is already
        running, connect to it and exit, otherwise run our functional
        default startup.
        """
        # Dispatch dbus CLI command
        if uri and not show_window:
            show_window = self.CLI_SHOW_MANAGER
        data = GLib.Variant("(sss)", (uri or "", show_window or "", domain or ""))

        is_remote = self._application.get_is_remote()
        if not is_remote:
            self._default_startup(skip_autostart, uri)
        self._application.activate_action("cli_command", data)

        if is_remote:
            log.debug("Connected to remote app instance.")
            return

        self._application.run(None)

    ###########################
    # timer and tick handling #
    ###########################

    def _timer_changed_cb(self, *args, **kwargs):
        ignore1 = args
        ignore2 = kwargs
        self._schedule_timer()

    def _schedule_timer(self):
        interval = self.config.get_stats_update_interval() * 1000

        if self._timer is not None:
            self.remove_gobject_timeout(self._timer)
            self._timer = None

        self._timer = self.timeout_add(interval, self._tick)

    def _add_obj_to_tick_queue(self, obj, isprio, **kwargs):
        if self._tick_queue.full():  # pragma: no cover
            if not self._tick_thread_slow:
                log.debug("Tick is slow, not running at requested rate.")
                self._tick_thread_slow = True
            return

        self._tick_counter += 1
        self._tick_queue.put((isprio and PRIO_HIGH or PRIO_LOW, self._tick_counter, obj, kwargs))

    def schedule_priority_tick(self, conn, kwargs):
        # Called directly from connection
        self._add_obj_to_tick_queue(conn, True, **kwargs)

    def _tick(self):
        for conn in self._connobjs.values():
            self._add_obj_to_tick_queue(conn, False, stats_update=True, pollvm=True)
        return 1

    def _handle_tick_queue(self):
        while True:
            ignore1, ignore2, conn, kwargs = self._tick_queue.get()
            try:
                conn.tick_from_engine(**kwargs)
            except Exception:  # pragma: no cover
                # Don't attempt to show any UI error here, since it
                # can cause dialogs to appear from nowhere if say
                # libvirtd is shut down
                log.debug("Error polling connection %s", conn.get_uri(), exc_info=True)

            # Need to clear reference to make leak check happy
            conn = None
            self._tick_queue.task_done()

    #####################################
    # window counting and exit handling #
    #####################################

    def increment_window_counter(self):
        """
        Public function, called by toplevel windows
        """
        self._window_count += 1
        log.debug("window counter incremented to %s", self._window_count)

    def decrement_window_counter(self):
        """
        Public function, called by toplevel windows
        """
        self._window_count -= 1
        log.debug("window counter decremented to %s", self._window_count)

        self._exit_app_if_no_windows()

    def _systray_is_embedded(self):
        """
        We don't use window tracking here: systray isn't a window and even
        when 'show' has been requested it may not be embedded in a visible
        tray area, so we have to check it separately.
        """
        return vmmSystray.get_instance().is_embedded()

    def _show_systray_from_cli(self):
        """
        Handler for --show-systray from CLI.
        We force show the systray, and wait for a timeout to report if
        its embedded. If not we raise an error and exit the app
        """
        vmmSystray.get_instance().show_from_cli()

        @_show_startup_error
        def check(self, count):
            count -= 1
            if self._systray_is_embedded():
                log.debug("systray embedded")
                return
            if count <= 0:  # pragma: no cover
                raise RuntimeError("systray did not show up")
            self.timeout_add(1000, check, self, count)  # pragma: no cover

        startcount = 5
        timeout = 1000
        if self.config.CLITestOptions.fake_systray:
            timeout = 1
        self.timeout_add(timeout, check, self, startcount)

    def _can_exit(self):
        return self._window_count <= 0 and not self._systray_is_embedded()

    def _exit_app_if_no_windows(self):
        if self._exiting:
            return
        if self._can_exit():
            log.debug("No windows found, requesting app exit")
            self.exit_app()

    def exit_app(self):
        """
        Public call, manager/details/... use this to force exit the app
        """
        if self._exiting:
            return  # pragma: no cover

        self._exiting = True

        def _do_exit():
            try:
                vmmConnectionManager.get_instance().cleanup()
                self.emit("app-closing")
                self.cleanup()

                if self.config.CLITestOptions.leak_debug:
                    objs = self.config.get_objects()
                    # Engine will always appear to leak
                    objs.remove(self.object_key)

                    for name in objs:  # pragma: no cover
                        log.debug("LEAK: %s", name)

                log.debug("Exiting app normally.")
            finally:
                self._application.quit()

        # We stick this in an idle callback, so the exit_app() caller
        # reference is dropped, and leak check debug doesn't give a
        # false positive
        self.idle_add(_do_exit)

    ##########################################
    # Window launchers from virt-manager cli #
    ##########################################

    def _find_vm_by_cli_str(self, uri, clistr):
        """
        Lookup a VM by a string passed in on the CLI. Can be either
        ID, domain name, or UUID
        """
        if clistr.isdigit():
            clistr = int(clistr)

        for vm in self._connobjs[uri].list_vms():
            if clistr == vm.get_id():
                return vm
            elif clistr == vm.get_name():
                return vm
            elif clistr == vm.get_uuid():
                return vm

    def _cli_show_vm_helper(self, uri, clistr, page):
        vm = self._find_vm_by_cli_str(uri, clistr)
        if not vm:
            raise RuntimeError("%s does not have VM '%s'" % (uri, clistr))

        from .vmwindow import vmmVMWindow

        details = vmmVMWindow.get_instance(None, vm)

        if page == self.CLI_SHOW_DOMAIN_PERFORMANCE:
            details.activate_performance_page()
        elif page in [self.CLI_SHOW_DOMAIN_EDITOR, self.CLI_SHOW_DOMAIN_DELETE]:
            details.activate_config_page()
        elif page == self.CLI_SHOW_DOMAIN_CONSOLE:
            details.activate_console_page()

        details.show()

        if page == self.CLI_SHOW_DOMAIN_DELETE:
            from .delete import vmmDeleteDialog

            vmmDeleteDialog.show_instance(details, vm)

    def _get_manager(self):
        from .manager import vmmManager

        return vmmManager.get_instance(None)

    @_show_startup_error
    def _launch_cli_window(self, uri, show_window, clistr):
        log.debug("Launching requested window '%s'", show_window)
        if show_window == self.CLI_SHOW_MANAGER:
            manager = self._get_manager()
            manager.set_initial_selection(uri)
            manager.show()
        elif show_window == self.CLI_SHOW_DOMAIN_CREATOR:
            from .createvm import vmmCreateVM

            vmmCreateVM.show_instance(None, uri)
        elif show_window == self.CLI_SHOW_HOST_SUMMARY:
            from .host import vmmHost

            vmmHost.show_instance(None, self._connobjs[uri])
        elif show_window in [
            self.CLI_SHOW_DOMAIN_EDITOR,
            self.CLI_SHOW_DOMAIN_PERFORMANCE,
            self.CLI_SHOW_DOMAIN_CONSOLE,
            self.CLI_SHOW_DOMAIN_DELETE,
        ]:
            self._cli_show_vm_helper(uri, clistr, show_window)
        elif show_window == self.CLI_SHOW_SYSTEM_TRAY:
            # Handled elsewhere
            pass
        else:  # pragma: no cover
            raise RuntimeError("Unknown cli window command '%s'" % show_window)

    def _handle_conn_error(self, _conn, ConnectError):
        msg, details, title = ConnectError
        modal = self._can_exit()
        self.err.show_err(msg, details, title, modal=modal)
        self._exit_app_if_no_windows()

    @_show_startup_error
    def _handle_cli_command(self, actionobj, variant):
        ignore = actionobj
        uri = variant[0]
        show_window = variant[1] or self.CLI_SHOW_MANAGER
        domain = variant[2]

        log.debug(
            "processing cli command uri=%s show_window=%s domain=%s", uri, show_window, domain
        )

        if show_window == self.CLI_SHOW_SYSTEM_TRAY:
            log.debug("Launching only in the system tray")
            self._show_systray_from_cli()
            if not uri:
                return
        elif not uri:
            log.debug("No cli action requested, launching default window")
            self._get_manager().show()
            return

        conn_is_new = uri not in self._connobjs
        conn = vmmConnectionManager.get_instance().add_conn(uri)
        if conn.is_active():
            self.idle_add(self._launch_cli_window, uri, show_window, domain)
            return

        def _open_completed(_c, ConnectError):
            if ConnectError:
                if conn_is_new:
                    log.debug("Removing failed uri=%s", uri)
                    vmmConnectionManager.get_instance().remove_conn(uri)
                self._handle_conn_error(conn, ConnectError)
            else:
                self._launch_cli_window(uri, show_window, domain)

        conn.connect_once("open-completed", _open_completed)
        conn.open()



# Copyright 2017 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


XMLBuilder.register_namespace("qemu", "http://libvirt.org/schemas/domain/qemu/1.0")


class _XMLNSQemuArg(XMLBuilder):
    XML_NAME = "qemu:arg"

    value = XMLProperty("./@value")


class _XMLNSQemuEnv(XMLBuilder):
    XML_NAME = "qemu:env"

    name = XMLProperty("./@name")
    value = XMLProperty("./@value")


class DomainXMLNSQemu(XMLBuilder):
    """
    Class for generating <qemu:commandline> XML
    """

    XML_NAME = "qemu:commandline"
    _XML_PROP_ORDER = ["args", "envs"]

    args = XMLChildProperty(_XMLNSQemuArg)
    envs = XMLChildProperty(_XMLNSQemuEnv)



# Copyright (c) 2018 Oracle and/or its affiliates. All rights reserved.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty, XMLChildProperty


###############
# CPU Pinning #
###############


class _VCPUPin(XMLBuilder):
    """
    Class for generating <cputune> child <vcpupin> XML
    """

    XML_NAME = "vcpupin"
    _XML_PROP_ORDER = ["vcpu", "cpuset"]

    vcpu = XMLProperty("./@vcpu", is_int=True)
    cpuset = XMLProperty("./@cpuset")


class _IOThreadPin(XMLBuilder):
    """
    Class for generating <cputune> child <iothreadpin> XML
    """

    XML_NAME = "iothreadpin"
    _XML_PROP_ORDER = ["iothread", "cpuset"]

    iothread = XMLProperty("./@iothread", is_int=True)
    cpuset = XMLProperty("./@cpuset")


##############
# Scheduling #
##############


class _VCPUSched(XMLBuilder):
    """
    Class for generating <cputune> child <vcpusched> XML
    """

    XML_NAME = "vcpusched"
    _XML_PROP_ORDER = ["vcpus", "scheduler", "priority"]

    vcpus = XMLProperty("./@vcpus")
    scheduler = XMLProperty("./@scheduler")
    priority = XMLProperty("./@priority", is_int=True)


class _IOThreadSched(XMLBuilder):
    """
    Class for generating <cputune> child <iothreadsched> XML
    """

    XML_NAME = "iothreadsched"
    _XML_PROP_ORDER = ["iothreads", "scheduler", "priority"]

    iothreads = XMLProperty("./@iothreads")
    scheduler = XMLProperty("./@scheduler")
    priority = XMLProperty("./@priority", is_int=True)


###########################
# Cache & Memory Tunables #
###########################


class _CacheTuneCache(XMLBuilder):
    """
    Class for generating <cachetune> child <cache> XML
    """

    XML_NAME = "cache"
    _XML_PROP_ORDER = ["id", "level", "type", "size", "unit"]

    id = XMLProperty("./@id", is_int=True)
    level = XMLProperty("./@level", is_int=True)
    type = XMLProperty("./@type")
    size = XMLProperty("./@size", is_int=True)
    unit = XMLProperty("./@unit")


class _CacheTuneMonitor(XMLBuilder):
    """
    Class for generating <cachetune> child <monitor> XML
    """

    XML_NAME = "monitor"
    _XML_PROP_ORDER = ["level", "vcpus"]

    level = XMLProperty("./@level", is_int=True)
    vcpus = XMLProperty("./@vcpus")


class _CacheTune(XMLBuilder):
    """
    Class for generating <cputune> child <cachetune> XML
    """

    XML_NAME = "cachetune"
    _XML_PROP_ORDER = ["vcpus", "caches"]

    vcpus = XMLProperty("./@vcpus")
    caches = XMLChildProperty(_CacheTuneCache)
    monitors = XMLChildProperty(_CacheTuneMonitor)


class _MemoryTuneNode(XMLBuilder):
    """
    Class for generating <memorytune> child <node> XML
    """

    XML_NAME = "node"
    _XML_PROP_ORDER = ["id", "bandwidth"]

    id = XMLProperty("./@id", is_int=True)
    bandwidth = XMLProperty("./@bandwidth", is_int=True)


class _MemoryTune(XMLBuilder):
    """
    Class for generating <cputune> child <memorytune> XML
    """

    XML_NAME = "memorytune"
    _XML_PROP_ORDER = ["vcpus", "nodes"]

    vcpus = XMLProperty("./@vcpus")
    nodes = XMLChildProperty(_MemoryTuneNode)


#########################
# Actual CPUTune domain #
#########################


class DomainCputune(XMLBuilder):
    """
    Class for generating <cputune> XML
    """

    XML_NAME = "cputune"
    _XML_PROP_ORDER = [
        "shares",
        "period",
        "quota",
        "global_period",
        "global_quota",
        "emulator_period",
        "emulator_quota",
        "iothread_period",
        "iothread_quota",
        "vcpupins",
        "emulatorpin_cpuset",
        "iothreadpins",
        "emulatorsched_scheduler",
        "emulatorsched_priority",
        "vcpuscheds",
        "iothreadscheds",
        "cachetunes",
        "memorytunes",
    ]

    # Resource quotas
    shares = XMLProperty("./shares", is_int=True)
    period = XMLProperty("./period", is_int=True)
    quota = XMLProperty("./quota", is_int=True)
    global_period = XMLProperty("./global_period", is_int=True)
    global_quota = XMLProperty("./global_quota", is_int=True)
    emulator_period = XMLProperty("./emulator_period", is_int=True)
    emulator_quota = XMLProperty("./emulator_quota", is_int=True)
    iothread_period = XMLProperty("./iothread_period", is_int=True)
    iothread_quota = XMLProperty("./iothread_quota", is_int=True)

    # CPU Pinning
    vcpupins = XMLChildProperty(_VCPUPin)
    emulatorpin_cpuset = XMLProperty("./emulatorpin/@cpuset")
    iothreadpins = XMLChildProperty(_IOThreadPin)

    # Scheduling
    emulatorsched_scheduler = XMLProperty("./emulatorsched/@scheduler")
    emulatorsched_priority = XMLProperty("./emulatorsched/@priority", is_int=True)
    vcpuscheds = XMLChildProperty(_VCPUSched)
    iothreadscheds = XMLChildProperty(_IOThreadSched)

    # Cache & Memory Tunables
    cachetunes = XMLChildProperty(_CacheTune)
    memorytunes = XMLChildProperty(_MemoryTune)



#
# Copyright 2006-2009, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import random

from .device import Device, DeviceAddress
from ..nodedev import NodeDevice
from ..logger import log
from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


def _random_mac(conn):
    """Generate a random MAC address.

    00-16-3E allocated to xensource
    52-54-00 used by qemu/kvm

    The OUI list is available at https://standards.ieee.org/regauth/oui/oui.txt.

    The remaining 3 fields are random, with the first bit of the first
    random field set 0.

    @return: MAC address string
    """

    if conn.is_qemu():
        oui = [0x52, 0x54, 0x00]
    else:
        # Xen
        oui = [0x00, 0x16, 0x3E]

    mac = oui + [random.randint(0x00, 0xFF), random.randint(0x00, 0xFF), random.randint(0x00, 0xFF)]
    return ":".join(["%02x" % x for x in mac])


def _default_route():
    route_file = "/proc/net/route"
    if not os.path.exists(route_file):  # pragma: no cover
        log.debug("route_file=%s does not exist", route_file)
        return None

    for line in open(route_file):
        info = line.split()
        if len(info) != 11:  # pragma: no cover
            log.debug("Unexpected field count=%s when parsing %s", len(info), route_file)
            break

        try:
            route = int(info[1], 16)
            if route == 0:
                return info[0]
        except ValueError:
            continue

    return None  # pragma: no cover


def _host_default_bridge():
    dev = _default_route()
    if not dev:
        return None  # pragma: no cover

    # New style peth0 == phys dev, eth0 == bridge, eth0 == default route
    if os.path.exists("/sys/class/net/%s/bridge" % dev):
        return dev  # pragma: no cover

    # Old style, peth0 == phys dev, eth0 == netloop, xenbr0 == bridge,
    # vif0.0 == netloop attached, eth0 == default route
    try:
        defn = int(dev[-1])
    except Exception:  # pragma: no cover
        defn = -1

    if (
        defn >= 0
        and os.path.exists("/sys/class/net/peth%d/brport" % defn)
        and os.path.exists("/sys/class/net/xenbr%d/bridge" % defn)
    ):
        return "xenbr%d"  # pragma: no cover
    return None


# Cache the host default bridge lookup. It can change over the lifetime
# of a virt-manager run, but that should be rare, and this saves us
# possibly spamming logs if host lookup goes wrong
_HOST_DEFAULT_BRIDGE = -1


def _default_bridge(conn):
    if conn.is_remote():
        return None

    global _HOST_DEFAULT_BRIDGE
    if _HOST_DEFAULT_BRIDGE == -1:
        try:
            ret = _host_default_bridge()
        except Exception:  # pragma: no cover
            log.debug("Error getting host default bridge", exc_info=True)
            ret = None
        _HOST_DEFAULT_BRIDGE = ret

    ret = _HOST_DEFAULT_BRIDGE
    if conn.in_testsuite():
        ret = "testsuitebr0"
    return ret


_MAC_COUNTER = 0


def _testsuite_mac():
    # Generate predictable mac addresses for the test suite
    # For some tests, we need to make sure that different mac addresses
    # would _not_ be generated in normal operations, so we add some magic
    # here to increment the generated address with a special env variable
    global _MAC_COUNTER

    base = "00:11:22:33:44:55"
    ret = base[:-1] + str(int(base[-1]) + _MAC_COUNTER)
    _MAC_COUNTER += 1

    if "VIRTINST_TEST_SUITE_INCREMENT_MACADDR" not in os.environ:
        _MAC_COUNTER = 0
    return ret


class _Backend(XMLBuilder):
    XML_NAME = "backend"

    type = XMLProperty("./@type")
    logFile = XMLProperty("./@logFile", do_abspath=True)


class _VirtualPort(XMLBuilder):
    XML_NAME = "virtualport"

    type = XMLProperty("./@type")
    managerid = XMLProperty("./parameters/@managerid", is_int=True)
    typeid = XMLProperty("./parameters/@typeid", is_int=True)
    typeidversion = XMLProperty("./parameters/@typeidversion", is_int=True)
    instanceid = XMLProperty("./parameters/@instanceid")
    profileid = XMLProperty("./parameters/@profileid")
    interfaceid = XMLProperty("./parameters/@interfaceid")


class _PortForwardRange(XMLBuilder):
    XML_NAME = "range"

    start = XMLProperty("./@start", is_int=True)
    end = XMLProperty("./@end", is_int=True)
    to = XMLProperty("./@to", is_int=True)
    exclude = XMLProperty("./@exclude", is_yesno=True)


class _PortForward(XMLBuilder):
    XML_NAME = "portForward"

    proto = XMLProperty("./@proto")
    address = XMLProperty("./@address")
    dev = XMLProperty("./@dev")

    range = XMLChildProperty(_PortForwardRange)


class _DeviceInterfaceSourceAddress(DeviceAddress):
    pass


class DeviceInterface(Device):
    XML_NAME = "interface"

    TYPE_BRIDGE = "bridge"
    TYPE_VIRTUAL = "network"
    TYPE_USER = "user"
    TYPE_VHOSTUSER = "vhostuser"
    TYPE_ETHERNET = "ethernet"
    TYPE_DIRECT = "direct"
    TYPE_VDPA = "vdpa"

    @staticmethod
    def generate_mac(conn):
        """
        Generate a random MAC that doesn't conflict with any VMs on
        the connection.
        """
        if conn.fake_conn_predictable():
            return _testsuite_mac()

        for ignore in range(256):
            mac = _random_mac(conn)
            try:
                DeviceInterface.check_mac_in_use(conn, mac)
                return mac
            except RuntimeError:  # pragma: no cover
                continue

        log.debug("Failed to generate non-conflicting MAC")  # pragma: no cover
        return None  # pragma: no cover

    @staticmethod
    def check_mac_in_use(conn, searchmac):
        """
        Raise RuntimeError if the passed mac conflicts with a defined VM
        """
        if not searchmac:
            return

        vms = conn.fetch_all_domains()
        for vm in vms:
            for nic in vm.devices.interface:
                nicmac = nic.macaddr or ""
                if nicmac.lower() == searchmac.lower():
                    raise RuntimeError(
                        _("The MAC address '%s' is in use by another virtual machine.") % searchmac
                    )

    @staticmethod
    def default_bridge(conn):
        """
        Return the bridge virt-install would use as a default value,
        if one is setup on the host
        """
        return _default_bridge(conn)

    ###############
    # XML helpers #
    ###############

    def _get_source(self):
        """
        Convenience function, try to return the relevant <source> value
        per the network type.
        """
        if self.type == self.TYPE_VIRTUAL:
            return self.network
        if self.type == self.TYPE_BRIDGE:
            return self.bridge
        if self.type == self.TYPE_DIRECT or self.type == self.TYPE_VDPA:
            return self.source_dev
        return None

    def _set_source(self, newsource):
        """
        Convenience function, try to set the relevant <source> value
        per the network type
        """
        self.bridge = None
        self.network = None
        self.source_dev = None

        if self.type == self.TYPE_VIRTUAL:
            self.network = newsource
        elif self.type == self.TYPE_BRIDGE:
            self.bridge = newsource
        elif self.type == self.TYPE_DIRECT or self.type == self.TYPE_VDPA:
            self.source_dev = newsource

    source = property(_get_source, _set_source)

    ##################
    # XML properties #
    ##################

    _XML_PROP_ORDER = [
        "backend",
        "bridge",
        "network",
        "source_dev",
        "source_type",
        "source_path",
        "source_mode",
        "portgroup",
        "macaddr",
        "target_dev",
        "model",
        "virtualport",
        "filterref",
        "rom_bar",
        "rom_file",
        "mtu_size",
        "portForward",
    ]

    backend = XMLChildProperty(_Backend, is_single=True)

    bridge = XMLProperty("./source/@bridge")
    network = XMLProperty("./source/@network")
    source_dev = XMLProperty("./source/@dev")

    virtualport = XMLChildProperty(_VirtualPort, is_single=True)
    type = XMLProperty("./@type")
    trustGuestRxFilters = XMLProperty("./@trustGuestRxFilters", is_yesno=True)

    macaddr = XMLProperty("./mac/@address")

    source_type = XMLProperty("./source/@type")
    source_path = XMLProperty("./source/@path")
    source_mode = XMLProperty("./source/@mode")
    source_address = XMLChildProperty(
        _DeviceInterfaceSourceAddress, is_single=True, relative_xpath="./source"
    )
    managed = XMLProperty("./@managed", is_yesno=True)

    portgroup = XMLProperty("./source/@portgroup")
    model = XMLProperty("./model/@type")
    target_dev = XMLProperty("./target/@dev")
    filterref = XMLProperty("./filterref/@filter")
    link_state = XMLProperty("./link/@state")

    driver_name = XMLProperty("./driver/@name")
    driver_queues = XMLProperty("./driver/@queues", is_int=True)

    rom_bar = XMLProperty("./rom/@bar", is_onoff=True)
    rom_file = XMLProperty("./rom/@file")

    mtu_size = XMLProperty("./mtu/@size", is_int=True)

    portForward = XMLChildProperty(_PortForward)

    #############
    # Build API #
    #############

    def set_default_source(self):
        if self.conn.is_qemu_unprivileged() or self.conn.is_test():
            self.type = self.TYPE_USER
            return

        nettype = DeviceInterface.TYPE_BRIDGE
        source = DeviceInterface.default_bridge(self.conn)
        if not source:
            nettype = DeviceInterface.TYPE_VIRTUAL
            source = "default"

        self.type = nettype
        self.source = source

    def set_from_nodedev(self, nodedev):
        log.debug("set_from_nodedev xml=\n%s", nodedev.get_xml())

        self.type = "hostdev"
        if self.managed is None:
            self.managed = True

        if nodedev.device_type == NodeDevice.CAPABILITY_TYPE_PCI:
            self.source_address.type = "pci"
            self.source_address.domain = nodedev.domain
            self.source_address.bus = nodedev.bus
            self.source_address.slot = nodedev.slot
            self.source_address.function = nodedev.function

        else:  # pragma: no cover
            raise ValueError(_("Unsupported node device type '%s'") % nodedev.device_type)

    ##################
    # Default config #
    ##################

    @staticmethod
    def default_model(guest):
        if not guest.os.is_hvm():
            return None
        if guest.supports_virtionet():
            return "virtio"
        if guest.os.is_q35():
            return "e1000e"
        if not guest.os.is_x86():
            return None

        prefs = ["e1000", "rtl8139", "ne2k_pci", "pcnet"]
        supported_models = guest.osinfo.supported_netmodels()
        for pref in prefs:
            if pref in supported_models:
                return pref
        return "e1000"

    def _set_default_type(self):
        if self.type:
            return

        if self.backend and self.backend.type == "passt":
            self.type = self.TYPE_USER
            return

        self.type = self.TYPE_BRIDGE

    def set_defaults(self, guest):
        self._set_default_type()

        if not self.macaddr:
            self.macaddr = self.generate_mac(self.conn)
        if self.type == self.TYPE_BRIDGE and not self.bridge:
            self.bridge = _default_bridge(self.conn)
        if not self.model:
            self.model = self.default_model(guest)



# Copyright (C) 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from gi.repository import Gtk

import virtinst
from virtinst import log

from ..lib import uiutil
from ..baseclass import vmmGObjectUI

(
    _EDIT_CACHE,
    _EDIT_DISCARD,
    _EDIT_RO,
    _EDIT_SHARE,
    _EDIT_REMOVABLE,
    _EDIT_SERIAL,
) = range(1, 7)


class vmmAddStorage(vmmGObjectUI):
    __gsignals__ = {
        "browse-clicked": (vmmGObjectUI.RUN_FIRST, None, [object]),
        "storage-toggled": (vmmGObjectUI.RUN_FIRST, None, [object]),
        "changed": (vmmGObjectUI.RUN_FIRST, None, []),
    }

    def __init__(self, conn, builder, topwin):
        vmmGObjectUI.__init__(self, "addstorage.ui", None, builder=builder, topwin=topwin)
        self.conn = conn

        def _e(edittype):
            def signal_cb(*args):
                self._change_cb(edittype)

            return signal_cb

        self.builder.connect_signals(
            {
                "on_storage_browse_clicked": self._browse_storage,
                "on_storage_select_toggled": self._toggle_storage_select,
                "on_disk_cache_combo_changed": _e(_EDIT_CACHE),
                "on_disk_discard_combo_changed": _e(_EDIT_DISCARD),
                "on_disk_readonly_changed": _e(_EDIT_RO),
                "on_disk_shareable_changed": _e(_EDIT_SHARE),
                "on_disk_removable_changed": _e(_EDIT_REMOVABLE),
                "on_disk_serial_changed": _e(_EDIT_SERIAL),
            }
        )

        self._active_edits = []
        self.top_box = self.widget("storage-box")
        self.advanced_top_box = self.widget("storage-advanced-box")
        self._init_ui()

    def _cleanup(self):
        self.conn = None
        self.top_box.destroy()
        self.advanced_top_box.destroy()

    ##########################
    # Initialization methods #
    ##########################

    def _host_disk_space(self):
        try:
            pool = self.conn.get_default_pool()
            if pool and pool.is_active():
                # Rate limit this, since it can be spammed at dialog startup time
                if pool.secs_since_last_refresh() > 10:
                    pool.refresh()
                avail = int(pool.get_available())
                return float(avail / 1024.0 / 1024.0 / 1024.0)
        except Exception:  # pragma: no cover
            log.exception("Error determining host disk space")
        return -1

    def _update_host_space(self):
        widget = self.widget("phys-hd-label")
        max_storage = self._host_disk_space()

        def pretty_storage(size):
            if size == -1:
                return "Unknown GiB"
            return "%.1f GiB" % float(size)

        hd_label = _("%s available in the default location") % pretty_storage(max_storage)
        hd_label = "<span>%s</span>" % hd_label
        widget.set_markup(hd_label)

    def _init_ui(self):
        # Disk cache combo
        values = [[None, _("Hypervisor default")]]
        for m in virtinst.DeviceDisk.CACHE_MODES:
            values.append([m, m])
        uiutil.build_simple_combo(self.widget("disk-cache"), values, sort=False)

        # Discard combo
        values = [[None, _("Hypervisor default")]]
        for m in virtinst.DeviceDisk.DISCARD_MODES:
            values.append([m, m])
        uiutil.build_simple_combo(self.widget("disk-discard"), values, sort=False)

    ##############
    # Public API #
    ##############

    @staticmethod
    def check_path_search(src, conn, path):
        skip_paths = src.config.get_perms_fix_ignore()
        searchdata = virtinst.DeviceDisk.check_path_search(conn.get_backend(), path)

        broken_paths = searchdata.fixlist[:]
        for p in broken_paths[:]:
            if p in skip_paths:
                broken_paths.remove(p)

        if not broken_paths:
            return

        log.debug("No search access for dirs: %s", broken_paths)
        resp, chkres = src.err.warn_chkbox(
            _("The emulator may not have search permissions for the path '%s'.") % path,
            _("Do you want to correct this now?"),
            _("Don't ask about these directories again."),
            buttons=Gtk.ButtonsType.YES_NO,
        )

        if chkres:
            src.config.add_perms_fix_ignore(broken_paths)
        if not resp:
            return

        log.debug("Attempting to correct permission issues.")
        errors = virtinst.DeviceDisk.fix_path_search(searchdata)
        if not errors:
            return

        errmsg = _("Errors were encountered changing permissions for the following directories:")
        details = ""
        for p, error in errors.items():
            if p in broken_paths:
                details += "%s : %s\n" % (p, error)
        details += "\nIt is very likely the VM will fail to start up."

        log.debug("Permission errors:\n%s", details)

        ignore, chkres = src.err.err_chkbox(
            errmsg, details, _("Don't ask about these directories again.")
        )

        if chkres:
            src.config.add_perms_fix_ignore(list(errors.keys()))

    def reset_state(self):
        self._update_host_space()
        self._active_edits = []
        self.widget("storage-create").set_active(True)
        self.widget("storage-size").set_value(20)
        self.widget("storage-entry").set_text("")
        self.widget("storage-create-box").set_sensitive(True)
        self.widget("disk-cache").set_active(0)
        self.widget("disk-discard").set_active(0)
        self.widget("disk-serial").set_text("")
        self.widget("storage-advanced").set_expanded(False)
        self.widget("disk-readonly").set_active(False)
        self.widget("disk-shareable").set_active(False)
        self.widget("disk-removable").set_active(False)
        uiutil.set_grid_row_visible(self.widget("disk-removable"), False)

        storage_tooltip = None

        can_storage = not self.conn.is_remote() or self.conn.support.conn_storage()
        use_storage = self.widget("storage-select")
        storage_area = self.widget("storage-box")

        storage_area.set_sensitive(can_storage)
        if not can_storage:  # pragma: no cover
            storage_tooltip = _("Connection does not support storage management.")
            use_storage.set_sensitive(True)
        storage_area.set_tooltip_text(storage_tooltip or "")

    def get_default_path(self, name, collideguest=None):
        pool = self.conn.get_default_pool()
        if not pool:
            return  # pragma: no cover

        fmt = self.conn.get_default_storage_format()
        suffix = virtinst.StorageVolume.get_file_extension_for_format(fmt)
        suffix = suffix or ".img"

        path = virtinst.StorageVolume.find_free_name(
            self.conn.get_backend(),
            pool.get_backend(),
            name,
            suffix=suffix,
            collideguest=collideguest,
        )

        return os.path.join(pool.xmlobj.target_path, path)

    def is_default_storage(self):
        return self.widget("storage-create").get_active()

    def build_device(self, vmname, path=None, device="disk", collideguest=None):
        if path is None:
            if self.is_default_storage():
                path = self.get_default_path(vmname, collideguest=collideguest)
            else:
                path = self.widget("storage-entry").get_text().strip()

        disk = virtinst.DeviceDisk(self.conn.get_backend())
        disk.set_source_path(path or None)
        disk.device = device
        vals = self.get_values()

        if vals.get("cache") is not None:
            disk.driver_cache = vals.get("cache")
        if vals.get("discard") is not None:
            disk.driver_discard = vals.get("discard")
        if vals.get("readonly") is not None:
            disk.read_only = vals.get("readonly")
        if vals.get("shareable") is not None:
            disk.shareable = vals.get("shareable")
        if vals.get("serial") is not None:
            disk.serial = vals.get("serial")
        if vals.get("removable") is not None and self.widget("disk-removable").get_visible():
            disk.removable = vals.get("removable")

        if disk.wants_storage_creation():
            path = disk.get_source_path()
            pool = disk.get_parent_pool()
            size = uiutil.spin_get_helper(self.widget("storage-size"))
            fmt = self.conn.get_default_storage_format()

            # If the user changed the default disk format to raw, assume
            # they want to maximize performance, so fully allocate the
            # disk image. Otherwise use sparse
            sparse = fmt != "raw"

            vol_install = virtinst.DeviceDisk.build_vol_install(
                disk.conn, os.path.basename(path), pool, size, sparse
            )
            disk.set_vol_install(vol_install)

            if disk.get_vol_install().supports_format():
                log.debug("Using default prefs format=%s for path=%s", fmt, path)
                disk.get_vol_install().format = fmt
            else:
                log.debug(
                    "path=%s can not use default prefs format=%s, not setting it", path, fmt
                )  # pragma: no cover

        return disk

    def validate_device(self, disk):
        if disk.is_empty() and disk.device in ["disk", "lun"]:
            return self.err.val_err(_("A storage path must be specified."))

        disk.validate()
        path = disk.get_source_path()

        # Disk collision
        names = disk.is_conflict_disk()
        if names:
            msg = _("Disk '%(path)s' is already in use by other guests %(names)s") % {
                "path": path,
                "names": names,
            }
            res = self.err.yes_no(msg, _("Do you really want to use the disk?"))
            if not res:
                return False

        self.check_path_search(self, self.conn, path)

    ##################
    # Device editing #
    ##################

    def set_disk_bus(self, bus):
        show_removable = bus == "usb"
        uiutil.set_grid_row_visible(self.widget("disk-removable"), show_removable)

    def set_dev(self, disk):
        cache = disk.driver_cache
        discard = disk.driver_discard
        ro = disk.read_only
        share = disk.shareable
        removable = bool(disk.removable)
        serial = disk.serial

        self.set_disk_bus(disk.bus)

        uiutil.set_list_selection(self.widget("disk-cache"), cache)
        uiutil.set_list_selection(self.widget("disk-discard"), discard)

        self.widget("disk-serial").set_text(serial or "")
        self.widget("disk-readonly").set_active(ro)
        self.widget("disk-readonly").set_sensitive(not disk.is_cdrom())
        self.widget("disk-shareable").set_active(share)
        self.widget("disk-removable").set_active(removable)

        # This comes last
        self._active_edits = []

    def get_values(self):
        ret = {}

        if _EDIT_CACHE in self._active_edits:
            ret["cache"] = uiutil.get_list_selection(self.widget("disk-cache"))
        if _EDIT_DISCARD in self._active_edits:
            ret["discard"] = uiutil.get_list_selection(self.widget("disk-discard"))
        if _EDIT_RO in self._active_edits:
            ret["readonly"] = self.widget("disk-readonly").get_active()
        if _EDIT_SHARE in self._active_edits:
            ret["shareable"] = self.widget("disk-shareable").get_active()
        if _EDIT_REMOVABLE in self._active_edits:
            ret["removable"] = bool(self.widget("disk-removable").get_active())
        if _EDIT_SERIAL in self._active_edits:
            ret["serial"] = self.widget("disk-serial").get_text()

        return ret

    #############
    # Listeners #
    #############

    def _browse_storage(self, ignore):
        self.emit("browse-clicked", self.widget("storage-entry"))

    def _toggle_storage_select(self, src):
        act = src.get_active()
        self.widget("storage-browse-box").set_sensitive(act)
        self.emit("storage-toggled", src)

    def _change_cb(self, edittype):
        if edittype not in self._active_edits:
            self._active_edits.append(edittype)
        self.emit("changed")



# Copyright (C) 2010, 2013 Red Hat, Inc.
# Copyright (C) 2010 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from virtinst import log
from virtinst import xmlutil

from ..baseclass import vmmGObject


class vmmLibvirtObject(vmmGObject):
    __gsignals__ = {
        "state-changed": (vmmGObject.RUN_FIRST, None, []),
        "initialized": (vmmGObject.RUN_FIRST, None, [bool]),
    }

    _STATUS_ACTIVE = 1
    _STATUS_INACTIVE = 2

    def __init__(self, conn, backend, name, parseclass):
        vmmGObject.__init__(self)
        self._conn = conn
        self._backend = backend
        self._parseclass = parseclass
        self._name = name

        self.__initialized = False
        self.__status = None

        self._xmlobj = None
        self._xmlobj_to_define = None
        self._is_xml_valid = False

        # These should be set by the child classes if necessary
        self._inactive_xml_flags = 0
        self._active_xml_flags = 0

    @staticmethod
    def log_redefine_xml_diff(obj, origxml, newxml):
        if origxml == newxml:
            log.debug("Redefine requested for %s, but XML didn't change!", obj)
            return

        diff = xmlutil.diff(origxml, newxml, "Original XML", "New XML")
        log.debug("Redefining %s with XML diff:\n%s", obj, diff)

    @staticmethod
    def lifecycle_action(fn):
        """
        Decorator for object lifecycle actions like start, stop, delete.
        Will make sure any necessary state is updated accordingly.
        """

        def newfn(self, *args, **kwargs):
            ret = fn(self, *args, **kwargs)

            # If events are supported, this is a no-op, but the event loop
            # will trigger force_status_update, which will refresh_xml as well.
            #
            # If events aren't supported, the priority tick will call
            # self.tick(), which will call force_status_update
            poll_param = self._conn_tick_poll_param()  # pylint: disable=protected-access
            tick_kwargs = {poll_param: True}
            self.conn.schedule_priority_tick(**tick_kwargs)

            return ret

        return newfn

    def __repr__(self):
        # pylint: disable=arguments-differ
        try:
            name = self.get_name()
        except Exception:
            name = ""
        return "<%s name=%s id=%s>" % (self.__class__.__name__, name, hex(id(self)))

    def _cleanup(self):
        self._backend = None

    def _get_conn(self):
        return self._conn

    conn = property(_get_conn)

    def get_backend(self):
        return self._backend

    def is_domain(self):
        return self.class_name() == "domain"

    def is_network(self):
        return self.class_name() == "network"

    def is_pool(self):
        return self.class_name() == "pool"

    def is_nodedev(self):
        return self.class_name() == "nodedev"

    def get_autostart(self):  # pragma: no cover
        return False

    def set_autostart(self, val):  # pragma: no cover
        ignore = val

    def change_name_backend(self, newbackend):
        # Used for changing the backing object after a rename
        self._backend = newbackend

    def define_name(self, newname):
        oldname = self.get_xmlobj().name
        oldautostart = self.get_autostart()

        self.ensure_latest_xml()
        xmlobj = self._make_xmlobj_to_define()
        if xmlobj.name == newname:
            return  # pragma: no cover

        log.debug("Changing %s name from %s to %s", self, oldname, newname)
        origxml = xmlobj.get_xml()
        xmlobj.name = newname
        newxml = xmlobj.get_xml()

        try:
            self._name = newname
            self.conn.rename_object(self, origxml, newxml)
        except Exception:  # pragma: no cover
            self._name = oldname
            raise
        finally:
            self.__force_refresh_xml()

        self.set_autostart(oldautostart)

    #############################################################
    # Functions that should probably be overridden in sub class #
    #############################################################

    def _XMLDesc(self, flags):
        raise NotImplementedError()

    def class_name(self):
        raise NotImplementedError()

    def _conn_tick_poll_param(self):
        # The parameter name for conn.tick() object polling. So
        # for vmmDomain == "pollvm"
        raise NotImplementedError()

    def reports_stats(self):
        return False

    def _using_events(self):
        return False

    def _get_backend_status(self):
        raise NotImplementedError()

    def _define(self, xml):  # pragma: no cover
        ignore = xml
        return

    def delete(self, force=True):  # pragma: no cover
        ignore = force

    def get_name(self):
        return self._name

    def tick(self, stats_update=True):
        ignore = stats_update
        self._refresh_status()

    def _init_libvirt_state(self):
        self.tick()

    def init_libvirt_state(self):
        """
        Function called by vmmConnection to populate initial state when
        a new object appears.
        """
        if self.__initialized:
            return  # pragma: no cover

        initialize_failed = False
        try:
            if self.config.CLITestOptions.object_denylist == self._name:
                raise RuntimeError("fake initialization error")

            self._init_libvirt_state()
        except Exception:  # pragma: no cover
            log.debug("Error initializing libvirt state for %s", self, exc_info=True)
            initialize_failed = True

        self.__initialized = True
        self.idle_emit("initialized", initialize_failed)

    ###################
    # Status handling #
    ###################

    def _get_status(self):
        return self.__status

    def is_active(self):
        # vmmDomain overwrites this since it has more fine grained statuses
        return self._get_status() == self._STATUS_ACTIVE

    def run_status(self):
        if self.is_active():
            return _("Active")
        return _("Inactive")

    def _refresh_status(self, newstatus=None, cansignal=True):
        """
        Grab the object status/active state from libvirt, and if the
        status has changed, update the XML cache. Typically called from
        object tick functions for manually updating the object state.

        :param newstatus: Used by vmmDomain as a small optimization to
            avoid polling info() twice
        :param cansignal: If True, this function will signal state-changed
            if required.
        :returns: True if status changed, false otherwise
        """
        if self._using_events() and self.__status is not None:
            return False

        if newstatus is None:
            newstatus = self._get_backend_status()
        status = newstatus
        if status == self.__status:
            return False
        self.__status = status

        self.ensure_latest_xml(nosignal=True)
        if cansignal:
            self.idle_emit("state-changed")
        return True

    ##################
    # Public XML API #
    ##################

    def recache_from_event_loop(self):
        """
        Updates the VM status and XML, because we received an event from
        libvirt's event implementations. That's the only time this should
        be used.

        We refresh status and XML because they are tied together in subtle
        ways, like runtime XML changing when a VM is started.
        """
        try:
            self.__force_refresh_xml(nosignal=True)
            # status = None forces a signal to be emitted
            self.__status = None
            self._refresh_status()
        except Exception as e:
            # If we hit an exception here, it's often that the object
            # disappeared, so request the poll loop to be updated
            log.debug("Error refreshing %s from events: %s", self, e)
            poll_param = self._conn_tick_poll_param()
            if poll_param:
                kwargs = {"force": True, poll_param: True}
                log.debug("Scheduling priority tick with: %s", kwargs)
                self.conn.schedule_priority_tick(**kwargs)

    def ensure_latest_xml(self, nosignal=False):
        """
        Refresh XML if it isn't up to date, basically if we aren't using
        events.
        """
        if self._using_events() and self._xmlobj and self._is_xml_valid:
            return
        self.__force_refresh_xml(nosignal=nosignal)

    def __force_refresh_xml(self, nosignal=False):
        """
        Force an xml update. Signal 'state-changed' if domain xml has
        changed since last refresh

        :param nosignal: If true, don't send state-changed. Used by
            callers that are going to send it anyways.
        """
        origxml = None
        if self._xmlobj:
            origxml = self._xmlobj.get_xml()

        self._invalidate_xml()
        active_xml = self._XMLDesc(self._active_xml_flags)
        self._xmlobj = self._parseclass(self.conn.get_backend(), parsexml=active_xml)
        self._is_xml_valid = True

        if not nosignal and origxml != active_xml:
            self.idle_emit("state-changed")

    def get_xmlobj(self, inactive=False, refresh_if_nec=True):
        """
        Get object xml, return it wrapped in a virtinst object.
        If cached xml is invalid, update.

        :param inactive: Return persistent XML, not the running config.
            No effect if domain is not running. Use this flag
            if the XML will be used for redefining a guest
        :param refresh_if_nec: Check if XML is out of date, and if so,
            refresh it (default behavior). Skipping a refresh is
            useful to prevent updating xml in the tick loop when
            it's not that important (disk/net stats)
        """
        if inactive:
            # If inactive XML requested, always return a fresh object even if
            # the current object is inactive XML (like when the domain is
            # stopped). Callers that request inactive are basically expecting
            # a new copy.
            inactive_xml = self._XMLDesc(self._inactive_xml_flags)
            return self._parseclass(self.conn.get_backend(), parsexml=inactive_xml)

        if self._xmlobj is None or (refresh_if_nec and not self._is_xml_valid):
            self.ensure_latest_xml()

        return self._xmlobj

    @property
    def xmlobj(self):
        return self.get_xmlobj()

    def get_xml_to_define(self):
        """
        Return the raw inactive XML we would use to alter/define an
        object. Used by the xmleditor UI
        """
        return self._make_xmlobj_to_define().get_xml()

    def define_xml(self, xml):
        """
        Define the passed in XML, and log a diff against the current XML.
        Generally subclasses should use _redefine_xmlobj with higher
        level wrappers, but this is needed for the XML editor
        """
        origxml = self.get_xml_to_define()
        newxml = xml
        self._redefine_xml_internal(origxml, newxml)

    #########################
    # Internal XML routines #
    #########################

    def _invalidate_xml(self):
        """
        Mark cached XML as invalid. Subclasses may extend this
        to invalidate any specific caches of their own
        """
        # While for events we do want to clear cached XML values like
        # _name, the XML is never invalid.
        self._is_xml_valid = self._using_events()

    def _make_xmlobj_to_define(self):
        """
        Build an xmlobj that should be used for defining new XML.

        Most subclasses shouldn't touch this, but vmmDomainVirtinst needs to.
        """
        return self.get_xmlobj(inactive=True)

    def _redefine_xml_internal(self, origxml, newxml):
        self.log_redefine_xml_diff(self, origxml, newxml)

        self._define(newxml)
        if self._using_events():
            return

        self.ensure_latest_xml(nosignal=True)
        self.idle_emit("state-changed")

    def _redefine_xmlobj(self, xmlobj):
        """
        Redefine the passed xmlobj, which should be generated with
        self._make_xmlobj_to_define() and which has accumulated edits
        from UI fields.

        Most subclasses shouldn't alter this, but vmmDomainVirtinst needs to.
        """
        origxml = self._make_xmlobj_to_define().get_xml()
        newxml = xmlobj.get_xml()
        self._redefine_xml_internal(origxml, newxml)



#
# Copyright 2020 Oracle Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceIommu(Device):
    XML_NAME = "iommu"

    model = XMLProperty("./@model")
    aw_bits = XMLProperty("./driver/@aw_bits", is_int=True)
    intremap = XMLProperty("./driver/@intremap", is_onoff=True)
    caching_mode = XMLProperty("./driver/@caching_mode", is_onoff=True)
    eim = XMLProperty("./driver/@eim", is_onoff=True)
    iotlb = XMLProperty("./driver/@iotlb", is_onoff=True)



#
# Copyright 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty


class DomainPm(XMLBuilder):
    XML_NAME = "pm"

    suspend_to_mem = XMLProperty("./suspend-to-mem/@enabled", is_yesno=True)
    suspend_to_disk = XMLProperty("./suspend-to-disk/@enabled", is_yesno=True)

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        # When the suspend feature is exposed to VMs, an ACPI shutdown
        # event triggers a suspend in the guest, which causes a lot of
        # user confusion (especially compounded with the face that suspend
        # is often buggy so VMs can get hung, etc).
        #
        # We've been disabling this in virt-manager for a while, but lets
        # do it here too for consistency.
        if guest.os.is_x86() and self.conn.support.conn_pm_disable():
            if self.suspend_to_mem is None:
                self.suspend_to_mem = False
            if self.suspend_to_disk is None:
                self.suspend_to_disk = False



from ..xmlbuilder import XMLBuilder, XMLProperty, XMLChildProperty


class _KeyWrap(XMLBuilder):

    XML_NAME = "cipher"
    _XML_PROP_ORDER = ["name", "state"]

    name = XMLProperty("./@name")
    state = XMLProperty("./@state", is_onoff=True)


class DomainKeyWrap(XMLBuilder):
    """
    Class for generating <keywrap> XML
    """

    XML_NAME = "keywrap"

    cipher = XMLChildProperty(_KeyWrap)



#
# Base class for all VM devices
#
# Copyright 2008, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import collections
import os
import re
import string
import textwrap

from .logger import log
from .xmlapi import XMLAPI
from . import xmlutil


# pylint: disable=protected-access
# This whole file is calling around into non-public functions that we
# don't want regular API users to touch

_trackprops = bool("VIRTINST_TEST_SUITE" in os.environ)
_allprops = []
_seenprops = []


class XMLManualAction:
    """
    Helper class for tracking and performing the user requested manual
    XML action
    """

    ACTION_CREATE = 1
    ACTION_DELETE = 2
    ACTION_SET = 3

    xpath_delete = None
    xpath_create = None
    xpath_value = None
    xpath_set = None

    def _process_args(self):
        def _ret(x, a, v=None):
            return (x, a, v)

        if self.xpath_delete:
            return _ret(self.xpath_delete, XMLManualAction.ACTION_DELETE)
        if self.xpath_create:
            return _ret(self.xpath_create, XMLManualAction.ACTION_CREATE)

        xpath = self.xpath_set
        if self.xpath_value:
            val = self.xpath_value
        else:
            if "=" not in str(xpath):
                raise ValueError("%s: Setting xpath must be in the form of XPATH=VALUE" % xpath)
            xpath, val = xpath.rsplit("=", 1)
        return _ret(xpath, XMLManualAction.ACTION_SET, val)

    def perform(self, xmlstate):
        if (
            not self.xpath_delete
            and not self.xpath_create
            and not self.xpath_value
            and not self.xpath_set
        ):
            return
        xpath, action, value = self._process_args()

        if xpath.startswith("."):
            xpath = xmlstate.make_abs_xpath(xpath)

        if action == self.ACTION_DELETE:
            setval = False
        elif action == self.ACTION_CREATE:
            setval = True
        else:
            setval = value or None
        xmlstate.xmlapi.set_xpath_content(xpath, setval)


class _XMLPropertyCache:
    """
    Cache lookup tables mapping classes to their associated
    XMLProperty and XMLChildProperty classes
    """

    def __init__(self):
        self._name_to_prop = {}
        self._prop_to_name = {}

    def _get_prop_cache(self, cls, checkclass):
        cachename = str(cls) + "-" + checkclass.__name__
        if cachename not in self._name_to_prop:
            ret = {}
            for c in reversed(type.mro(cls)[:-1]):
                for key, val in c.__dict__.items():
                    if isinstance(val, checkclass):
                        ret[key] = val
                        self._prop_to_name[val] = key
            self._name_to_prop[cachename] = ret
        return self._name_to_prop[cachename]

    def get_xml_props(self, inst):
        return self._get_prop_cache(inst.__class__, XMLProperty)

    def get_child_props(self, inst):
        return self._get_prop_cache(inst.__class__, XMLChildProperty)

    def get_prop_name(self, propinst):
        return self._prop_to_name[propinst]


_PropCache = _XMLPropertyCache()


class _XMLChildList(list):
    """
    Little wrapper for a list containing XMLChildProperty output.
    This is just to insert a dynamically created add_new() function
    which instantiates and appends a new child object
    """

    def __init__(self, childclass, copylist, xmlbuilder, is_xml=True):
        list.__init__(self)
        self._childclass = childclass
        self._xmlbuilder = xmlbuilder
        self._is_xml = is_xml
        for i in copylist:
            self.append(i)

    def new(self):
        """
        Instantiate a new child object and return it
        """
        args = ()
        if self._is_xml:
            args = (self._xmlbuilder.conn,)
        return self._childclass(*args)

    def add_new(self):
        """
        Instantiate a new child object, append it, and return it
        """
        obj = self.new()
        if self._is_xml:
            self._xmlbuilder.add_child(obj)
        else:
            self.append(obj)
        return obj


class _XMLPropertyBase(property):
    def __init__(self, fget, fset):
        self._propname = None
        property.__init__(self, fget=fget, fset=fset)

    @property
    def propname(self):
        """
        The variable name associated with this XMLProperty. So with
        a definition like

            foo = XMLProperty("./@bar")

        and this will return "foo".
        """
        if not self._propname:
            self._propname = _PropCache.get_prop_name(self)
        return self._propname


class XMLChildProperty(_XMLPropertyBase):
    """
    Property that points to a class used for parsing a subsection of
    of the parent XML. For example when we delegate parsing
    /domain/cpu/feature of the /domain/cpu class.

    @child_class: XMLBuilder class this property is tracking. So for
        guest.devices.disk this is DeviceDisk
    @relative_xpath: Relative location where the class is rooted compared
        to its xmlbuilder root path. So if xmlbuilder is ./foo and we
        want to track ./foo/bar/baz instances, set relative_xpath=./bar
    @is_single: If True, this represents an XML node that is only expected
        to appear once, like <domain><cpu>
    """

    def __init__(self, child_class, relative_xpath=".", is_single=False):
        self.child_class = child_class
        self.is_single = is_single
        self.relative_xpath = relative_xpath

        _XMLPropertyBase.__init__(self, self._fget, None)

    def __repr__(self):
        return "<XMLChildProperty %s %s>" % (str(self.child_class), id(self))

    def _get(self, xmlbuilder):
        if self.propname not in xmlbuilder._propstore and not self.is_single:
            xmlbuilder._propstore[self.propname] = []
        return xmlbuilder._propstore[self.propname]

    def _fget(self, xmlbuilder):
        if self.is_single:
            return self._get(xmlbuilder)
        return _XMLChildList(self.child_class, self._get(xmlbuilder), xmlbuilder)

    def clear(self, xmlbuilder):
        if self.is_single:
            self._get(xmlbuilder).clear()
        else:
            for obj in self._get(xmlbuilder)[:]:
                xmlbuilder.remove_child(obj)

    def insert(self, xmlbuilder, newobj, idx):
        self._get(xmlbuilder).insert(idx, newobj)

    def append(self, xmlbuilder, newobj):
        self._get(xmlbuilder).append(newobj)

    def remove(self, xmlbuilder, obj):
        self._get(xmlbuilder).remove(obj)

    def set(self, xmlbuilder, obj):
        xmlbuilder._propstore[self.propname] = obj

    def get_prop_xpath(self, _xmlbuilder, obj):
        return self.relative_xpath + "/" + obj.XML_NAME


class XMLProperty(_XMLPropertyBase):
    def __init__(
        self, xpath, is_bool=False, is_int=False, is_yesno=False, is_onoff=False, do_abspath=False
    ):
        """
        Set a XMLBuilder class property that maps to a value in an XML
        document, indicated by the passed xpath. For example, for a
        <domain><name> the definition may look like:

          name = XMLProperty("./name")

        When building XML from scratch (virt-install), 'name' works
        similar to a regular class property(). When parsing and editing
        existing guest XML, we  use the xpath value to get/set the value
        in the parsed XML document.

        :param xpath: xpath string which maps to the associated property
                      in a typical XML document
        :param name: Just a string to print for debugging, only needed
            if xpath isn't specified.
        :param is_bool: Whether this is a boolean property in the XML
        :param is_int: Whether this is an integer property in the XML
        :param is_yesno: Whether this is a yes/no property in the XML
        :param is_onoff: Whether this is an on/off property in the XML
        :param do_abspath: If True, run os.path.abspath on the passed value
        """
        self._xpath = xpath
        if not self._xpath:
            raise xmlutil.DevError("XMLProperty: xpath must be passed.")

        self._is_bool = is_bool
        self._is_int = is_int
        self._is_yesno = is_yesno
        self._is_onoff = is_onoff
        self._do_abspath = do_abspath

        conflicts = sum(
            int(bool(i)) for i in [self._is_bool, self._is_int, self._is_yesno, self._is_onoff]
        )
        if conflicts > 1:
            raise xmlutil.DevError("Conflict property converter options.")

        self._is_tracked = False
        if _trackprops:
            _allprops.append(self)

        _XMLPropertyBase.__init__(self, self.getter, self.setter)

    def __repr__(self):
        return "<XMLProperty %s %s>" % (str(self._xpath), id(self))

    ####################
    # Internal helpers #
    ####################

    def _convert_get_value(self, val):
        # pylint: disable=redefined-variable-type
        if self._is_bool:
            return bool(val)
        elif self._is_int and val is not None:
            try:
                intkwargs = {}
                if "0x" in str(val):
                    intkwargs["base"] = 16
                ret = int(val, **intkwargs)
            except ValueError as e:
                log.debug("Error converting XML value to int: %s", e)
                ret = val
        elif self._is_yesno:
            if val == "yes":
                ret = True
            elif val == "no":
                ret = False
            else:
                ret = val
        elif self._is_onoff:
            if val == "on":
                ret = True
            elif val == "off":
                ret = False
            else:
                ret = val
        else:
            ret = val
        return ret

    def _convert_set_value(self, val):
        if self._do_abspath and val is not None:
            val = os.path.abspath(val)
        elif self._is_onoff:
            if val is True:
                val = "on"
            elif val is False:
                val = "off"
        elif self._is_yesno:
            if val is True:
                val = "yes"
            elif val is False:
                val = "no"
        elif self._is_int and val is not None:
            intkwargs = {}
            if "0x" in str(val):
                intkwargs["base"] = 16
            val = int(val, **intkwargs)
        return val

    def _nonxml_fset(self, xmlbuilder, val):
        """
        This stores the value in XMLBuilder._propstore
        dict as propname->value. This saves us from having to explicitly
        track every variable.
        """
        propstore = xmlbuilder._propstore

        if self.propname in propstore:
            del propstore[self.propname]
        propstore[self.propname] = val

    def _nonxml_fget(self, xmlbuilder):
        """
        The flip side to nonxml_fset, fetch the value from
        XMLBuilder._propstore
        """
        return xmlbuilder._propstore.get(self.propname, None)

    def clear(self, xmlbuilder):
        # We only unset the cached data, since XML will be cleared elsewhere
        propstore = xmlbuilder._propstore
        if self.propname in propstore:
            self.setter(xmlbuilder, None)

    ##################################
    # The actual getter/setter impls #
    ##################################

    def getter(self, xmlbuilder):
        """
        Fetch value at user request. If we are parsing existing XML and
        the user hasn't done a 'set' yet, return the value from the XML,
        otherwise return the value from propstore

        If this is a built from scratch object, we never pull from XML
        since it's known to the empty, and we may want to return
        a 'default' value
        """
        if _trackprops and not self._is_tracked:
            _seenprops.append(self)
            self._is_tracked = True

        if self.propname in xmlbuilder._propstore:
            val = self._nonxml_fget(xmlbuilder)
        else:
            val = self._get_xml(xmlbuilder)
        return self._convert_get_value(val)

    def _get_xml(self, xmlbuilder):
        """
        Actually fetch the associated value from the backing XML
        """
        xpath = xmlbuilder._xmlstate.make_abs_xpath(self._xpath)
        return xmlbuilder._xmlstate.xmlapi.get_xpath_content(xpath, self._is_bool)

    def setter(self, xmlbuilder, val):
        """
        Set the value at user request. This just stores the value
        in propstore. Setting the actual XML is only done at
        get_xml time.
        """
        if _trackprops and not self._is_tracked:
            _seenprops.append(self)
            self._is_tracked = True

        setval = self._convert_set_value(val)
        self._nonxml_fset(xmlbuilder, setval)

    def _set_xml(self, xmlbuilder, setval):
        """
        Actually set the passed value in the XML document
        """
        xpath = xmlbuilder._xmlstate.make_abs_xpath(self._xpath)
        xmlbuilder._xmlstate.xmlapi.set_xpath_content(xpath, setval)


class _XMLState:
    def __init__(self, root_name, parsexml, parentxmlstate, relative_object_xpath):
        self._root_name = root_name
        self._namespace = ""
        if ":" in self._root_name:
            ns = self._root_name.split(":")[0]
            self._namespace = " xmlns:%s='%s'" % (ns, XMLAPI.NAMESPACES[ns])

        # xpath of this object relative to its parent. So for a standalone
        # <disk> this is empty, but if the disk is the forth one in a <domain>
        # it will be set to ./devices/disk[4]
        self._relative_object_xpath = relative_object_xpath or ""

        # xpath of the parent. For a disk in a standalone <domain>, this
        # is empty, but if the <domain> is part of a <domainsnapshot>,
        # it will be "./domain"
        self._parent_xpath = (parentxmlstate and parentxmlstate.abs_xpath()) or ""

        self.xmlapi = None
        self.is_build = not parsexml and not parentxmlstate
        self.parse(parsexml, parentxmlstate)

    def parse(self, parsexml, parentxmlstate):
        if parentxmlstate:
            self.is_build = parentxmlstate.is_build or self.is_build
            self.xmlapi = parentxmlstate.xmlapi
            return

        # Make sure passed in XML has required xmlns inserted
        if not parsexml:
            parsexml = "<%s%s/>" % (self._root_name, self._namespace)
        elif self._namespace and "xmlns" not in parsexml:
            parsexml = parsexml.replace(
                "<" + self._root_name, "<" + self._root_name + self._namespace
            )

        try:
            self.xmlapi = XMLAPI(parsexml)
        except Exception:
            log.debug("Error parsing xml=\n%s", parsexml)
            raise

        if not self.is_build:
            # Ensure parsexml has the correct root node
            self.xmlapi.validate_root_name(self._root_name.split(":")[-1])

    def set_relative_object_xpath(self, xpath):
        self._relative_object_xpath = xpath or ""

    def set_parent_xpath(self, xpath):
        self._parent_xpath = xpath or ""

    def _join_xpath(self, x1, x2):
        if x2.startswith("."):
            x2 = x2[1:]
        return x1 + x2

    def abs_xpath(self):
        return self._join_xpath(self._parent_xpath or ".", self._relative_object_xpath or ".")

    def make_abs_xpath(self, xpath):
        """
        Convert a relative xpath to an absolute xpath. So for DeviceDisk
        that's part of a Guest, accessing driver_name will do convert:
            ./driver/@name
        to an absolute xpath like:
            ./devices/disk[3]/driver/@name
        """
        return self._join_xpath(self.abs_xpath() or ".", xpath)


class XMLBuilder:
    """
    Base for all classes which build or parse domain XML
    """

    # Order that we should apply values to the XML. Keeps XML generation
    # consistent with what the test suite expects.
    _XML_PROP_ORDER = []

    # Name of the root XML element
    XML_NAME = None

    # In some cases, libvirt can incorrectly generate unparsable XML.
    # These are libvirt bugs, but this allows us to work around it in
    # for specific XML classes.
    #
    # Example: nodedev 'system' XML:
    # https://bugzilla.redhat.com/show_bug.cgi?id=1184131
    _XML_SANITIZE = False

    @staticmethod
    def register_namespace(nsname, uri):
        XMLAPI.register_namespace(nsname, uri)

    @staticmethod
    def validate_generic_name(name_label, val):
        # Only character that shouldn't work is '/', matching QEMU
        forbid = ["/"]
        if not val:
            # translators: value is a generic object type name
            raise ValueError(_("A name must be specified for the %s") % name_label)
        for c in forbid:
            if c not in val:
                continue
            msg = _("%(objecttype)s name '%(name)s' can not contain '%(char)s' character.") % {
                "objecttype": name_label,
                "name": val,
                "char": c,
            }
            raise ValueError(msg)

    def __init__(self, conn, parsexml=None, parentxmlstate=None, relative_object_xpath=None):
        """
        Initialize state

        :param conn: VirtinstConnection to validate device against
        :param parsexml: Optional XML string to parse

        The rest of the parameters are for internal use only
        """
        self.conn = conn

        if self._XML_SANITIZE:
            parsexml = parsexml.encode("ascii", "ignore").decode("ascii")
            parsexml = "".join([c for c in parsexml if c in string.printable])

        self._propstore = collections.OrderedDict()
        self._xmlstate = _XMLState(self.XML_NAME, parsexml, parentxmlstate, relative_object_xpath)

        self._validate_xmlbuilder()
        self._initial_child_parse()
        self.xml_actions = _XMLChildList(XMLManualAction, [], self, is_xml=False)

    def _validate_xmlbuilder(self):
        # This is one time validation we run once per XMLBuilder class
        cachekey = self.__class__.__name__ + "_xmlbuilder_validated"
        if getattr(self.__class__, cachekey, False):
            return

        xmlprops = self._all_xml_props()
        childprops = self._all_child_props()
        for key in self._XML_PROP_ORDER:
            if key not in xmlprops and key not in childprops:
                raise xmlutil.DevError("key '%s' must be xml prop or child prop" % key)

        childclasses = []
        for childprop in childprops.values():
            if childprop.child_class in childclasses:
                raise xmlutil.DevError(
                    "can't register duplicate child_class=%s" % childprop.child_class
                )
            childclasses.append(childprop.child_class)

        setattr(self.__class__, cachekey, True)

    def _initial_child_parse(self):
        # Walk the XML tree and hand of parsing to any registered
        # child classes
        for xmlprop in list(self._all_child_props().values()):
            child_class = xmlprop.child_class
            prop_path = xmlprop.get_prop_xpath(self, child_class)

            if xmlprop.is_single:
                obj = child_class(
                    self.conn, parentxmlstate=self._xmlstate, relative_object_xpath=prop_path
                )
                xmlprop.set(self, obj)
                continue

            nodecount = self._xmlstate.xmlapi.count(self._xmlstate.make_abs_xpath(prop_path))
            for idx in range(nodecount):
                idxstr = "[%d]" % (idx + 1)
                obj = child_class(
                    self.conn,
                    parentxmlstate=self._xmlstate,
                    relative_object_xpath=(prop_path + idxstr),
                )
                xmlprop.append(self, obj)

    def __repr__(self):
        return "<%s %s %s>" % (self.__class__.__name__.split(".")[-1], self.XML_NAME, id(self))

    ############################
    # Public XML managing APIs #
    ############################

    def get_xml(self):
        """
        Return XML string of the object
        """
        xmlapi = self._xmlstate.xmlapi
        if self._xmlstate.is_build:
            xmlapi = xmlapi.copy_api()

        self._add_parse_bits(xmlapi)
        ret = xmlapi.get_xml(self._xmlstate.make_abs_xpath("."))

        if not ret:
            return ret

        lastline = ret.rstrip().splitlines()[-1]
        if not ret.startswith(" ") and lastline.startswith(" "):
            ret = lastline.split("<")[0] + ret

        if not ret.endswith("\n"):
            ret += "\n"
        return ret

    def clear(self, leave_stub=False):
        """
        Wipe out all properties of the object

        :param leave_stub: if True, don't unlink the top stub node,
            see virtinst/cli usage for an explanation
        """
        props = list(self._all_xml_props().values())
        props += list(self._all_child_props().values())
        for prop in props:
            prop.clear(self)

        is_child = bool(re.match(r"^.*\[\d+\]$", self._xmlstate.abs_xpath()))
        if is_child or leave_stub:
            # User requested to clear an object that is the child of
            # another object (xpath ends in [1] etc). We can't fully remove
            # the node in that case, since then the xmlbuilder object is
            # no longer valid, and all the other child xpaths will be
            # pointing to the wrong node. So just stub out the content
            self._xmlstate.xmlapi.node_clear(self._xmlstate.abs_xpath())
        else:
            self._xmlstate.xmlapi.node_force_remove(self._xmlstate.abs_xpath())

    def validate(self):
        """
        Validate any set values and raise an exception if there's
        a problem
        """

    def set_defaults(self, guest):
        """
        Encode any default values if needed
        """
        ignore = guest

    def get_xml_id(self):
        """
        Return the location of the object in the XML document. This is
        basically the absolute xpath, but the value returned should be
        treated as opaque, it's just for cross XML comparisons. Used
        in virt-manager code
        """
        return self._xmlstate.abs_xpath()

    def get_xml_idx(self):
        """
        This is basically the offset parsed out of the object's xpath,
        minus 1. So if this is the fifth <disk> in a <domain>, ret=4.
        If this is the only <cpu> in a domain, ret=0.
        """
        xpath = self._xmlstate.abs_xpath()
        if "[" not in xpath:
            return 0
        return int(xpath.rsplit("[", 1)[1].strip("]")) - 1

    ################
    # Internal API #
    ################

    def _all_xml_props(self):
        """
        Return a list of all XMLProperty instances that this class has.
        """
        return _PropCache.get_xml_props(self)

    def _all_child_props(self):
        """
        Return a list of all XMLChildProperty instances that this class has.
        """
        return _PropCache.get_child_props(self)

    def _find_child_prop(self, child_class):
        xmlprops = self._all_child_props()
        ret = None
        for xmlprop in list(xmlprops.values()):
            if xmlprop.is_single:
                continue
            if child_class is xmlprop.child_class:
                ret = xmlprop
                break
        if not ret:
            raise xmlutil.DevError("Didn't find child property for child_class=%s" % child_class)
        return ret

    def _set_xpaths(self, parent_xpath, relative_object_xpath=-1):
        """
        Change the object hierarchy's cached xpaths
        """
        self._xmlstate.set_parent_xpath(parent_xpath)
        if relative_object_xpath != -1:
            self._xmlstate.set_relative_object_xpath(relative_object_xpath)
        for propname in self._all_child_props():
            for p in xmlutil.listify(getattr(self, propname, [])):
                p._set_xpaths(self._xmlstate.abs_xpath())

    def _set_child_xpaths(self):
        """
        Walk the list of child properties and make sure their
        xpaths point at their particular element. Needs to be called
        whenever child objects are added or removed
        """
        typecount = {}
        for propname, xmlprop in self._all_child_props().items():
            for obj in xmlutil.listify(getattr(self, propname)):
                idxstr = ""
                if not xmlprop.is_single:
                    class_type = obj.__class__
                    if class_type not in typecount:
                        typecount[class_type] = 0
                    typecount[class_type] += 1
                    idxstr = "[%d]" % typecount[class_type]

                prop_path = xmlprop.get_prop_xpath(self, obj)
                obj._set_xpaths(self._xmlstate.abs_xpath(), prop_path + idxstr)

    def _parse_with_children(self, *args, **kwargs):
        """
        Set new backing XML objects in ourselves and all our child props
        """
        self._xmlstate.parse(*args, **kwargs)
        for propname in self._all_child_props():
            for p in xmlutil.listify(getattr(self, propname, [])):
                p._parse_with_children(None, self._xmlstate)

    def add_child(self, obj, idx=None):
        """
        Insert the passed XMLBuilder object into our XML document. The
        object needs to have an associated mapping via XMLChildProperty
        """
        xmlprop = self._find_child_prop(obj.__class__)
        xml = obj.get_xml()
        if idx is None:
            xmlprop.append(self, obj)
        else:
            xmlprop.insert(self, obj, idx)
        self._set_child_xpaths()

        # Only insert the XML directly into the parent XML for !is_build
        # This is the only way to dictate XML ordering when building
        # from scratch, otherwise elements appear in the order they
        # are set. It's just a style issue but annoying nonetheless
        if not obj._xmlstate.is_build:
            use_xpath = obj._xmlstate.abs_xpath().rsplit("/", 1)[0]
            indent = 2 * obj._xmlstate.abs_xpath().count("/")
            self._xmlstate.xmlapi.node_add_xml(textwrap.indent(xml, indent * " "), use_xpath)
        obj._parse_with_children(None, self._xmlstate)

    def remove_child(self, obj):
        """
        Remove the passed XMLBuilder object from our XML document, but
        ensure its data isn't altered.
        """
        xmlprop = self._find_child_prop(obj.__class__)
        xmlprop.remove(self, obj)

        xpath = obj._xmlstate.abs_xpath()
        xml = obj.get_xml()
        obj._set_xpaths(None, None)
        obj._parse_with_children(xml, None)
        self._xmlstate.xmlapi.node_force_remove(xpath)
        self._set_child_xpaths()

    def replace_child(self, origobj, newobj):
        """
        Replace the origobj child with the newobj. For is_build, this
        replaces the objects, but for !is_build this only replaces the
        XML and keeps the object references in place. This is hacky and
        it's fixable but at time or writing it doesn't matter for
        our usecases.
        """
        if not self._xmlstate.is_build:
            xpath = origobj.get_xml_id()
            indent = 2 * xpath.count("/")
            xml = textwrap.indent(newobj.get_xml(), indent * " ").strip()
            self._xmlstate.xmlapi.node_replace_xml(xpath, xml)
        else:
            origidx = origobj.get_xml_idx()
            self.remove_child(origobj)
            self.add_child(newobj, idx=origidx)

    def _prop_is_unset(self, propname):
        """
        Return True if the property name has never had a value set
        """
        if getattr(self, propname):
            return False
        return propname not in self._propstore

    #################################
    # Private XML building routines #
    #################################

    def _add_parse_bits(self, xmlapi):
        """
        Callback that adds the implicitly tracked XML properties to
        the backing xml.
        """
        origpropstore = self._propstore.copy()
        origapi = self._xmlstate.xmlapi
        try:
            self._xmlstate.xmlapi = xmlapi
            return self._do_add_parse_bits()
        finally:
            self._xmlstate.xmlapi = origapi
            self._propstore = origpropstore

    def _do_add_parse_bits(self):
        # Set all defaults if the properties have one registered
        xmlprops = self._all_xml_props()
        childprops = self._all_child_props()

        # Set up preferred XML ordering
        do_order = [p for p in self._propstore if p not in childprops]
        for key in reversed(self._XML_PROP_ORDER):
            if key in do_order:
                do_order.remove(key)
                do_order.insert(0, key)
            elif key in childprops:
                do_order.insert(0, key)

        for key in sorted(list(childprops.keys())):
            if key not in do_order:
                do_order.append(key)

        # Alter the XML
        for key in do_order:
            if key in xmlprops:
                xmlprops[key]._set_xml(self, self._propstore[key])
            elif key in childprops:
                for obj in xmlutil.listify(getattr(self, key)):
                    obj._add_parse_bits(self._xmlstate.xmlapi)

        for manualaction in self.xml_actions:
            manualaction.perform(self._xmlstate)



# Copyright (C) 2006, 2012-2015 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from virtinst import DomainCpu
from virtinst import log

from .lib.inspection import vmmInspection


CSSDATA = """
/* Lighter colored text in some wizard summary fields */
.vmm-lighter {
    color: @insensitive_fg_color
}

/* Text on the blue header in our wizards */
.vmm-header-text {
    color: white
}

/* Subtext on the blue header in our wizards */
.vmm-header-subtext {
    color: #59B0E2
}

/* The blue header */
.vmm-header {
    background-color: #0072A8
}
"""


class _SettingsWrapper:
    """
    Wrapper class to simplify interacting with gsettings APIs.
    Basically it allows simple get/set of gconf style paths, and
    we internally convert it to the settings nested hierarchy. Makes
    client code much smaller.
    """

    def __init__(self, settings_id, gsettings_keyfile):
        self._root = settings_id

        if gsettings_keyfile:
            backend = Gio.keyfile_settings_backend_new(gsettings_keyfile, "/")
        else:
            backend = Gio.SettingsBackend.get_default()

        self._settings = Gio.Settings.new_with_backend(self._root, backend)

        self._settingsmap = {"": self._settings}
        self._handler_map = {}

        for child in self._settings.list_children():
            childschema = self._root + "." + child
            self._settingsmap[child] = Gio.Settings.new_with_backend(childschema, backend)

    ###################
    # Private helpers #
    ###################

    def _parse_key(self, key):
        value = key.strip("/")
        settingskey = ""
        if "/" in value:
            settingskey, value = value.rsplit("/", 1)
        return settingskey, value

    def _find_settings(self, key):
        settingskey, value = self._parse_key(key)
        return self._settingsmap[settingskey], value

    ###############
    # Public APIs #
    ###############

    def make_vm_settings(self, key):
        """
        Initialize per-VM relocatable schema if necessary
        """
        settingskey = self._parse_key(key)[0]
        if settingskey in self._settingsmap:
            return True

        schema = self._root + ".vm"
        path = "/" + self._root.replace(".", "/") + key.rsplit("/", 1)[0] + "/"
        self._settingsmap[settingskey] = Gio.Settings.new_with_path(schema, path)
        return True

    def make_conn_settings(self, key):
        """
        Initialize per-conn relocatable schema if necessary
        """
        settingskey = self._parse_key(key)[0]
        if settingskey in self._settingsmap:
            return True

        schema = self._root + ".connection"
        path = "/" + self._root.replace(".", "/") + key.rsplit("/", 1)[0] + "/"
        self._settingsmap[settingskey] = Gio.Settings.new_with_path(schema, path)
        return True

    def notify_add(self, key, cb, *args, **kwargs):
        settings, key = self._find_settings(key)

        def wrapcb(*ignore):
            return cb(*args, **kwargs)

        ret = settings.connect("changed::%s" % key, wrapcb, *args, **kwargs)
        self._handler_map[ret] = settings
        return ret

    def notify_remove(self, h):
        settings = self._handler_map.pop(h)
        return settings.disconnect(h)

    def get(self, key):
        settings, key = self._find_settings(key)
        return settings.get_value(key).unpack()

    def set(self, key, value, *args, **kwargs):
        settings, key = self._find_settings(key)
        fmt = settings.get_value(key).get_type_string()
        return settings.set_value(key, GLib.Variant(fmt, value), *args, **kwargs)


class vmmConfig:
    CONSOLE_SCALE_NEVER = 0
    CONSOLE_SCALE_FULLSCREEN = 1
    CONSOLE_SCALE_ALWAYS = 2

    _instance = None

    @classmethod
    def get_instance(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = vmmConfig(*args, **kwargs)
        return cls._instance

    @classmethod
    def is_initialized(cls):
        return bool(cls._instance)

    def __init__(self, BuildConfig, CLITestOptions):
        self.appname = "virt-manager"
        self.appversion = BuildConfig.version
        self.conf_dir = "/org/virt-manager/%s/" % self.appname
        self.ui_dir = BuildConfig.ui_dir

        self.conf = _SettingsWrapper(
            "org.virt-manager.virt-manager", CLITestOptions.gsettings_keyfile
        )

        self.CLITestOptions = CLITestOptions
        if self.CLITestOptions.xmleditor_enabled:
            self.set_xmleditor_enabled(True)
        if self.CLITestOptions.enable_libguestfs:
            self.set_libguestfs_inspect_vms(True)
        if self.CLITestOptions.disable_libguestfs:
            self.set_libguestfs_inspect_vms(False)

        # We don't create it straight away, since we don't want
        # to block the app pending user authorization to access
        # the keyring
        self._keyring = None

        self.default_graphics_from_config = BuildConfig.default_graphics
        self.default_hvs = BuildConfig.default_hvs

        self.default_storage_format_from_config = "qcow2"
        self.default_console_resizeguest = 0

        self._objects = []
        self.color_insensitive = None
        self._init_css()

    def _init_css(self):
        from gi.repository import Gdk

        screen = Gdk.Screen.get_default()

        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(CSSDATA.encode("utf-8"))

        context = Gtk.StyleContext()
        context.add_provider_for_screen(screen, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)

        found, color = context.lookup_color("insensitive_fg_color")
        if not found:  # pragma: no cover
            log.debug("Didn't find insensitive_fg_color in theme")
            return
        self.color_insensitive = color.to_string()

    # General app wide helpers (gsettings agnostic)

    def get_appname(self):
        return self.appname

    def get_appversion(self):
        return self.appversion

    def get_ui_dir(self):
        return self.ui_dir

    def inspection_supported(self):
        if not vmmInspection.libguestfs_installed():
            return False  # pragma: no cover
        return self.get_libguestfs_inspect_vms()

    def remove_notifier(self, h):
        self.conf.notify_remove(h)

    # Used for debugging reference leaks, we keep track of all objects
    # come and go so we can do a leak report at app shutdown
    def add_object(self, obj):
        self._objects.append(obj)

    def remove_object(self, obj):
        self._objects.remove(obj)

    def get_objects(self):
        return self._objects[:]

    #####################################
    # Wrappers for setting per-VM value #
    #####################################

    def _make_pervm_key(self, uuid, key):
        return "/vms/%s%s" % (uuid.replace("-", ""), key)

    def listen_pervm(self, uuid, key, *args, **kwargs):
        key = self._make_pervm_key(uuid, key)
        self.conf.make_vm_settings(key)
        return self.conf.notify_add(key, *args, **kwargs)

    def set_pervm(self, uuid, key, *args, **kwargs):
        key = self._make_pervm_key(uuid, key)
        self.conf.make_vm_settings(key)
        ret = self.conf.set(key, *args, **kwargs)
        return ret

    def get_pervm(self, uuid, key):
        key = self._make_pervm_key(uuid, key)
        self.conf.make_vm_settings(key)
        return self.conf.get(key)

    ########################################
    # Wrappers for setting per-conn values #
    ########################################

    def _make_perconn_key(self, uri, key):
        return "/conns/%s%s" % (uri.replace("/", ""), key)

    def listen_perconn(self, uri, key, *args, **kwargs):
        key = self._make_perconn_key(uri, key)
        self.conf.make_conn_settings(key)
        return self.conf.notify_add(key, *args, **kwargs)

    def set_perconn(self, uri, key, *args, **kwargs):
        key = self._make_perconn_key(uri, key)
        self.conf.make_conn_settings(key)
        ret = self.conf.set(key, *args, **kwargs)
        return ret

    def get_perconn(self, uri, key):
        key = self._make_perconn_key(uri, key)
        self.conf.make_conn_settings(key)
        return self.conf.get(key)

    ###################
    # General helpers #
    ###################

    # Manager stats view preferences
    def is_vmlist_guest_cpu_usage_visible(self):
        return self.conf.get("/vmlist-fields/cpu-usage")

    def is_vmlist_host_cpu_usage_visible(self):
        return self.conf.get("/vmlist-fields/host-cpu-usage")

    def is_vmlist_memory_usage_visible(self):
        return self.conf.get("/vmlist-fields/memory-usage")

    def is_vmlist_disk_io_visible(self):
        return self.conf.get("/vmlist-fields/disk-usage")

    def is_vmlist_network_traffic_visible(self):
        return self.conf.get("/vmlist-fields/network-traffic")

    def set_vmlist_guest_cpu_usage_visible(self, state):
        self.conf.set("/vmlist-fields/cpu-usage", state)

    def set_vmlist_host_cpu_usage_visible(self, state):
        self.conf.set("/vmlist-fields/host-cpu-usage", state)

    def set_vmlist_memory_usage_visible(self, state):
        self.conf.set("/vmlist-fields/memory-usage", state)

    def set_vmlist_disk_io_visible(self, state):
        self.conf.set("/vmlist-fields/disk-usage", state)

    def set_vmlist_network_traffic_visible(self, state):
        self.conf.set("/vmlist-fields/network-traffic", state)

    def on_vmlist_guest_cpu_usage_visible_changed(self, cb):
        return self.conf.notify_add("/vmlist-fields/cpu-usage", cb)

    def on_vmlist_host_cpu_usage_visible_changed(self, cb):
        return self.conf.notify_add("/vmlist-fields/host-cpu-usage", cb)

    def on_vmlist_memory_usage_visible_changed(self, cb):
        return self.conf.notify_add("/vmlist-fields/memory-usage", cb)

    def on_vmlist_disk_io_visible_changed(self, cb):
        return self.conf.notify_add("/vmlist-fields/disk-usage", cb)

    def on_vmlist_network_traffic_visible_changed(self, cb):
        return self.conf.notify_add("/vmlist-fields/network-traffic", cb)

    # Keys preferences
    def get_keys_combination(self):
        ret = self.conf.get("/console/grab-keys")
        if not ret:
            # Left Control + Left Alt
            return "65507,65513"
        return ret

    def set_keys_combination(self, val):
        # Val have to be a list of integers
        val = ",".join([str(v) for v in val])
        self.conf.set("/console/grab-keys", val)

    def on_keys_combination_changed(self, cb):
        return self.conf.notify_add("/console/grab-keys", cb)

    # Confirmation preferences
    def get_confirm_forcepoweroff(self):
        return self.conf.get("/confirm/forcepoweroff")

    def get_confirm_poweroff(self):
        return self.conf.get("/confirm/poweroff")

    def get_confirm_pause(self):
        return self.conf.get("/confirm/pause")

    def get_confirm_removedev(self):
        return self.conf.get("/confirm/removedev")

    def get_confirm_unapplied(self):
        return self.conf.get("/confirm/unapplied-dev")

    def get_confirm_delstorage(self):
        return self.conf.get("/confirm/delete-storage")

    def set_confirm_forcepoweroff(self, val):
        self.conf.set("/confirm/forcepoweroff", val)

    def set_confirm_poweroff(self, val):
        self.conf.set("/confirm/poweroff", val)

    def set_confirm_pause(self, val):
        self.conf.set("/confirm/pause", val)

    def set_confirm_removedev(self, val):
        self.conf.set("/confirm/removedev", val)

    def set_confirm_unapplied(self, val):
        self.conf.set("/confirm/unapplied-dev", val)

    def set_confirm_delstorage(self, val):
        self.conf.set("/confirm/delete-storage", val)

    # System tray visibility
    def on_view_system_tray_changed(self, cb):
        return self.conf.notify_add("/system-tray", cb)

    def get_view_system_tray(self):
        return self.conf.get("/system-tray")

    def set_view_system_tray(self, val):
        self.conf.set("/system-tray", val)

    # XML editor enabled
    def on_xmleditor_enabled_changed(self, cb):
        return self.conf.notify_add("/xmleditor-enabled", cb)

    def get_xmleditor_enabled(self):
        return self.conf.get("/xmleditor-enabled")

    def set_xmleditor_enabled(self, val):
        self.conf.set("/xmleditor-enabled", val)

    # Libguestfs VM inspection
    def get_libguestfs_inspect_vms(self):
        return self.conf.get("/enable-libguestfs-vm-inspection")

    def set_libguestfs_inspect_vms(self, val):
        self.conf.set("/enable-libguestfs-vm-inspection", val)

    # Stats history and interval length
    def get_stats_history_length(self):
        return 120

    def get_stats_update_interval(self):
        if self.CLITestOptions.short_poll:
            return 0.1
        interval = self.conf.get("/stats/update-interval")
        return max(interval, 1)

    def set_stats_update_interval(self, interval):
        self.conf.set("/stats/update-interval", interval)

    def on_stats_update_interval_changed(self, cb):
        return self.conf.notify_add("/stats/update-interval", cb)

    # Disable/Enable different stats polling
    def get_stats_enable_cpu_poll(self):
        return self.conf.get("/stats/enable-cpu-poll")

    def get_stats_enable_disk_poll(self):
        return self.conf.get("/stats/enable-disk-poll")

    def get_stats_enable_net_poll(self):
        return self.conf.get("/stats/enable-net-poll")

    def get_stats_enable_memory_poll(self):
        return self.conf.get("/stats/enable-memory-poll")

    def set_stats_enable_cpu_poll(self, val):
        self.conf.set("/stats/enable-cpu-poll", val)

    def set_stats_enable_disk_poll(self, val):
        self.conf.set("/stats/enable-disk-poll", val)

    def set_stats_enable_net_poll(self, val):
        self.conf.set("/stats/enable-net-poll", val)

    def set_stats_enable_memory_poll(self, val):
        self.conf.set("/stats/enable-memory-poll", val)

    def on_stats_enable_cpu_poll_changed(self, cb, row=None):
        return self.conf.notify_add("/stats/enable-cpu-poll", cb, row)

    def on_stats_enable_disk_poll_changed(self, cb, row=None):
        return self.conf.notify_add("/stats/enable-disk-poll", cb, row)

    def on_stats_enable_net_poll_changed(self, cb, row=None):
        return self.conf.notify_add("/stats/enable-net-poll", cb, row)

    def on_stats_enable_memory_poll_changed(self, cb, row=None):
        return self.conf.notify_add("/stats/enable-memory-poll", cb, row)

    def get_console_scaling(self):
        return self.conf.get("/console/scaling")

    def set_console_scaling(self, pref):
        self.conf.set("/console/scaling", pref)

    def get_console_resizeguest(self):
        val = self.conf.get("/console/resize-guest")
        if val == -1:
            val = self.default_console_resizeguest
        return val

    def set_console_resizeguest(self, pref):
        self.conf.set("/console/resize-guest", pref)

    def get_auto_usbredir(self):
        return bool(self.conf.get("/console/auto-redirect"))

    def set_auto_usbredir(self, state):
        self.conf.set("/console/auto-redirect", state)

    def get_console_autoconnect(self):
        return bool(self.conf.get("/console/autoconnect"))

    def set_console_autoconnect(self, val):
        return self.conf.set("/console/autoconnect", val)

    # Show VM details toolbar
    def get_details_show_toolbar(self):
        res = self.conf.get("/details/show-toolbar")
        if res is None:
            res = True  # pragma: no cover
        return res

    def set_details_show_toolbar(self, state):
        self.conf.set("/details/show-toolbar", state)

    # New VM preferences
    def get_graphics_type(self, raw=False):
        ret = self.conf.get("/new-vm/graphics-type")
        if ret not in ["system", "vnc", "spice"]:
            ret = "system"  # pragma: no cover
        if ret == "system" and not raw:
            return self.default_graphics_from_config
        return ret

    def set_graphics_type(self, gtype):
        self.conf.set("/new-vm/graphics-type", gtype.lower())

    def get_default_storage_format(self, raw=False):
        ret = self.conf.get("/new-vm/storage-format")
        if ret not in ["default", "raw", "qcow2"]:
            ret = "default"  # pragma: no cover
        if ret == "default" and not raw:
            return self.default_storage_format_from_config
        return ret

    def set_storage_format(self, typ):
        self.conf.set("/new-vm/storage-format", typ.lower())

    def get_default_cpu_setting(self):
        ret = self.conf.get("/new-vm/cpu-default")

        if ret not in DomainCpu.SPECIAL_MODES:
            ret = DomainCpu.SPECIAL_MODE_APP_DEFAULT  # pragma: no cover
        return ret

    def set_default_cpu_setting(self, val):
        self.conf.set("/new-vm/cpu-default", val.lower())

    def get_default_firmware_setting(self):
        ret = self.conf.get("/new-vm/firmware")
        if ret not in ["default", "uefi"]:
            ret = "default"  # pragma: no cover
        return ret

    def set_firmware_setting(self, val):
        self.conf.set("/new-vm/firmware", val.lower())

    # URL/Media path history
    def _url_add_helper(self, gsettings_path, url):
        maxlength = 10
        urls = self.conf.get(gsettings_path) or []

        if urls.count(url) == 0 and len(url) > 0 and not url.isspace():
            # The url isn't already in the list, so add it
            urls.insert(0, url)
            if len(urls) > maxlength:
                del urls[len(urls) - 1]  # pragma: no cover
            self.conf.set(gsettings_path, urls)

    def add_container_url(self, url):
        self._url_add_helper("/urls/containers", url)

    def get_container_urls(self):
        return self.conf.get("/urls/containers") or []

    def add_media_url(self, url):
        self._url_add_helper("/urls/urls", url)

    def get_media_urls(self):
        return self.conf.get("/urls/urls") or []

    def add_iso_path(self, path):
        self._url_add_helper("/urls/isos", path)

    def get_iso_paths(self):
        return self.conf.get("/urls/isos") or []

    def on_iso_paths_changed(self, cb):
        return self.conf.notify_add("/urls/isos", cb)

    # Whether to ask about fixing path permissions
    def add_perms_fix_ignore(self, pathlist):
        current_list = self.get_perms_fix_ignore() or []
        for path in pathlist:
            if path in current_list:
                continue  # pragma: no cover
            current_list.append(path)
        self.conf.set("/paths/perms-fix-ignore", current_list)

    def get_perms_fix_ignore(self):
        return self.conf.get("/paths/perms-fix-ignore")

    # Manager view connection list
    def get_conn_uris(self):
        return self.conf.get("/connections/uris") or []

    def add_conn_uri(self, uri):
        uris = self.get_conn_uris()
        if uri not in uris:
            uris.insert(len(uris) - 1, uri)
            self.conf.set("/connections/uris", uris)

    def remove_conn_uri(self, uri):
        uris = self.get_conn_uris()
        if uri in uris:
            uris.remove(uri)
            self.conf.set("/connections/uris", uris)

        if self.get_conn_autoconnect(uri):
            uris = self.conf.get("/connections/autoconnect")
            uris.remove(uri)
            self.conf.set("/connections/autoconnect", uris)

    # Manager default window size
    def get_manager_window_size(self):
        w = self.conf.get("/manager-window-width")
        h = self.conf.get("/manager-window-height")
        return (w, h)

    def set_manager_window_size(self, w, h):
        self.conf.set("/manager-window-width", w)
        self.conf.set("/manager-window-height", h)

    # URI autoconnect
    def get_conn_autoconnect(self, uri):
        uris = self.conf.get("/connections/autoconnect")
        return (uris is not None) and (uri in uris)

    def set_conn_autoconnect(self, uri, val):
        uris = self.conf.get("/connections/autoconnect") or []
        if not val and uri in uris:
            uris.remove(uri)
        elif val and uri not in uris:
            uris.append(uri)

        self.conf.set("/connections/autoconnect", uris)

    # Default directory location dealings
    def get_default_directory(self, gsettings_key):
        path = self.conf.get("/paths/%s-default" % gsettings_key)
        log.debug("directory for gsettings_key=%s returning=%s", gsettings_key, path)
        return path

    def set_default_directory(self, gsettings_key, folder):
        if not folder or folder.startswith("/dev"):
            return  # pragma: no cover
        log.debug("saving directory for gsettings_key=%s to %s", gsettings_key, folder)
        self.conf.set("/paths/%s-default" % gsettings_key, folder)
